"""
Functions for plotting datasets nicely.
"""
# TODO: custom xtick labels                                                   #
# TODO: annotations, arbitrary text                                           #
# TODO: docs                                                                  #

import functools
import numpy as np
from ..manage import auto_xyz_ds
from .core import (
    Plotter,
    AbstractLinePlot,
    AbstractScatter,
    AbstractHistogram,
    AbstractHeatMap,
    PLOTTER_DEFAULTS,
    calc_row_col_datasets,
    intercept_call_arg,
    prettify,
)
from .color import xyz_colormaps


# ----------------- Main lineplot interface for matplotlib ------------------ #

class PlotterMatplotlib(Plotter):
    """
    """

    def __init__(self, ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
        super().__init__(ds, x, y, z=z, y_err=y_err, x_err=x_err,
                         **kwargs, backend='MATPLOTLIB')

    def prepare_axes(self):
        """
        """
        import matplotlib as mpl
        if self.math_serif:
            mpl.rcParams['mathtext.fontset'] = 'cm'
            mpl.rcParams['mathtext.rm'] = 'serif'
        mpl.rcParams['font.family'] = self.font
        import matplotlib.pyplot as plt

        if self.axes_rloc is not None:
            if self.axes_loc is not None:
                raise ValueError("Cannot specify absolute and relative "
                                 "location of axes at the same time.")
            if self.add_to_fig is None:
                raise ValueError("Can only specify relative axes position "
                                 "when adding to a figure, i.e. when "
                                 "add_to_fig != None")

        if self.axes_rloc is not None:
            self._axes_loc = self._cax_rel2abs_rect(
                self.axes_rloc, self.add_to_fig.get_axes()[-1])
        else:
            self._axes_loc = self.axes_loc

        # Add a new set of axes to an existing plot
        if self.add_to_fig is not None and self.subplot is None:
            self._fig = self.add_to_fig
            self._axes = self._fig.add_axes((0.4, 0.6, 0.30, 0.25)
                                            if self._axes_loc is None else
                                            self._axes_loc)

        # Add lines to an existing set of axes
        elif self.add_to_axes is not None:
            self._fig = self.add_to_axes
            self._axes = self._fig.get_axes()[-1]

        # Add lines to existing axes but only sharing the x-axis
        elif self.add_to_xaxes is not None:
            self._fig = self.add_to_xaxes
            self._axes = self._fig.get_axes()[-1].twinx()

        elif self.subplot is not None:
            # Add new axes as subplot to existing subplot
            if self.add_to_fig is not None:
                self._fig = self.add_to_fig

            # New figure but add as subplot
            else:
                self._fig = plt.figure(self.fignum, figsize=self.figsize,
                                       dpi=100)
            self._axes = self._fig.add_subplot(self.subplot)

        # Make new figure and axes
        else:
            self._fig = plt.figure(self.fignum, figsize=self.figsize, dpi=100)
            self._axes = self._fig.add_axes((0.15, 0.15, 0.8, 0.75)
                                            if self._axes_loc is None else
                                            self._axes_loc)
        self._axes.set_title("" if self.title is None else self.title,
                             fontsize=self.fontsize_title)

    def set_axes_labels(self):
        if self._xtitle:
            self._axes.set_xlabel(self._xtitle, fontsize=self.fontsize_xtitle)
            self._axes.xaxis.labelpad = self.xtitle_pad
        if self._ytitle:
            self._axes.set_ylabel(self._ytitle, fontsize=self.fontsize_ytitle)
            self._axes.yaxis.labelpad = self.ytitle_pad
            if self.ytitle_right:
                self._axes.yaxis.set_label_position("right")

    def set_axes_scale(self):
        """
        """
        self._axes.set_xscale("log" if self.xlog else "linear")
        self._axes.set_yscale("log" if self.ylog else "linear")

    def set_axes_range(self):
        """
        """
        if self._xlims:
            self._axes.set_xlim(self._xlims)
        if self._ylims:
            self._axes.set_ylim(self._ylims)

    def set_spans(self):
        """
        """
        if self.vlines is not None:
            for x in self.vlines:
                self._axes.axvline(x, lw=self.span_width,
                                   color=self.span_color,
                                   linestyle=self.span_style)
        if self.hlines is not None:
            for y in self.hlines:
                self._axes.axhline(y, lw=self.span_width,
                                   color=self.span_color,
                                   linestyle=self.span_style)

    def set_gridlines(self):
        """
        """
        for axis in ('top', 'bottom', 'left', 'right'):
            self._axes.spines[axis].set_linewidth(1.0)

        if self.gridlines:
            # matplotlib has coarser gridine style than bokeh
            self._gridline_style = [x / 2 for x in self.gridline_style]
            self._axes.set_axisbelow(True)  # ensures gridlines below all
            self._axes.grid(True, color="0.9", which='major',
                            linestyle=(0, self._gridline_style))
            self._axes.grid(True, color="0.95", which='minor',
                            linestyle=(0, self._gridline_style))

    def set_tick_marks(self):
        """
        """
        import matplotlib as mpl

        if self.xticks is not None:
            self._axes.set_xticks(self.xticks, minor=False)
            self._axes.get_xaxis().set_major_formatter(
                mpl.ticker.ScalarFormatter())
        if self.yticks is not None:
            self._axes.set_yticks(self.yticks, minor=False)
            self._axes.get_yaxis().set_major_formatter(
                mpl.ticker.ScalarFormatter())

        if self.xtick_labels is not None:
            self._axes.set_xticklabels(self.xtick_labels)

        if self.xticklabels_hide:
            (self._axes.get_xaxis()
             .set_major_formatter(mpl.ticker.NullFormatter()))
        if self.yticklabels_hide:
            (self._axes.get_yaxis()
             .set_major_formatter(mpl.ticker.NullFormatter()))

        self._axes.tick_params(labelsize=self.fontsize_ticks, direction='out',
                               bottom='bottom' in self.ticks_where,
                               top='top' in self.ticks_where,
                               left='left' in self.ticks_where,
                               right='right' in self.ticks_where)

        if self.yticklabels_right or (self.yticklabels_right is None and
                                      self.ytitle_right is True):
            self._axes.yaxis.tick_right()

    def _cax_rel2abs_rect(self, rel_rect, cax=None):
        """Turn a relative axes specification into a absolute one.
        """
        if cax is None:
            cax = self._axes
        bbox = cax.get_position()
        l, b, w, h = bbox.x0, bbox.y0, bbox.width, bbox.height
        cl = l + w * rel_rect[0]
        cb = b + h * rel_rect[1]
        try:
            cw = w * rel_rect[2]
            ch = h * rel_rect[3]
        except IndexError:
            return cl, cb
        return cl, cb, cw, ch

    def plot_legend(self, grid=False, labels_handles=None):
        """Add a legend
        """
        if self._use_legend:

            if labels_handles:
                labels, handles = zip(*labels_handles.items())
            else:
                handles, labels = self._legend_handles, self._legend_labels

            if self.legend_reverse:
                handles, labels = handles[::-1], labels[::-1]

            # Limit minimum size of markers that appear in legend
            should_auto_scale_legend_markers = (
                (self.legend_marker_scale is None) and  # not already set
                hasattr(self, '_markersize') and  # is a valid parameter
                self._markersize < 3  # and is small
            )
            if should_auto_scale_legend_markers:
                self.legend_marker_scale = 3 / self._markersize

            opts = {
                'title': (self.z_coo if self.ztitle is None else self.ztitle),
                'loc': self.legend_loc,
                'fontsize': self.fontsize_zlabels,
                'frameon': self.legend_frame,
                'numpoints': 1,
                'scatterpoints': 1,
                'handlelength': self.legend_handlelength,
                'markerscale': self.legend_marker_scale,
                'labelspacing': self.legend_label_spacing,
                'columnspacing': self.legend_column_spacing,
                'bbox_to_anchor': self.legend_bbox,
                'ncol': self.legend_ncol
            }

            if grid:
                bb = opts['bbox_to_anchor']
                if bb is None:
                    opts['bbox_to_anchor'] = (1, 0.5, 0, 0)
                    opts['loc'] = 'center left'
                else:
                    loc = opts['loc']
                    # will get warning for 'best'
                    opts['loc'] = 'center' if loc in ('best', 0) else loc
                lgnd = self._fig.legend(handles, labels, **opts)
            else:
                lgnd = self._axes.legend(handles, labels, **opts)

            lgnd.get_title().set_fontsize(self.fontsize_ztitle)

            if self.legend_marker_alpha is not None:
                for l in lgnd.legendHandles:
                    l.set_alpha(1.0)

    def set_mappable(self):
        """Mappale object for colorbars.
        """
        from matplotlib.cm import ScalarMappable
        self.mappable = ScalarMappable(cmap=self.cmap, norm=self._color_norm)
        self.mappable.set_array([])

    def plot_colorbar(self, grid=False):
        """Add a colorbar to the data.
        """

        if self._use_colorbar:
            # Whether the colorbar should clip at either end
            extendmin = (self.vmin is not None) and (self.vmin > self._zmin)
            extendmax = (self.vmax is not None) and (self.vmax < self._zmax)
            extend = ('both' if extendmin and extendmax else
                      'min' if extendmin else
                      'max' if extendmax else
                      'neither')

            opts = {'extend': extend, 'ticks': self.zticks}

            if self.colorbar_relative_position:
                opts['cax'] = self._fig.add_axes(
                    self._cax_rel2abs_rect(self.colorbar_relative_position))

            if grid:
                opts['ax'] = self._fig.axes
                opts['anchor'] = (0.5, 0.5)

            self._cbar = self._fig.colorbar(
                self.mappable, **opts, **self.colorbar_opts)

            self._cbar.ax.tick_params(labelsize=self.fontsize_zlabels)

            self._cbar.ax.set_title(
                self._ctitle, fontsize=self.fontsize_ztitle,
                color=self.colorbar_color if self.colorbar_color else None)

            if self.colorbar_color:
                self._cbar.ax.yaxis.set_tick_params(
                    color=self.colorbar_color, labelcolor=self.colorbar_color)
                self._cbar.outline.set_edgecolor(self.colorbar_color)

    def set_panel_label(self):
        if self.panel_label is not None:
            self._axes.text(*self.panel_label_loc, self.panel_label,
                            transform=self._axes.transAxes,
                            fontsize=self.fontsize_panel_label,
                            color=self.panel_label_color,
                            ha='left', va='top')

    def show(self):
        import matplotlib.pyplot as plt
        if self.return_fig:
            plt.close(self._fig)
            return self._fig

    def prepare_plot(self):
        """Do all the things that every plot has.
        """
        self.prepare_axes()
        self.set_axes_labels()
        self.set_axes_scale()
        self.set_axes_range()
        self.set_spans()
        self.set_gridlines()
        self.set_tick_marks()

# --------------------------------------------------------------------------- #


def mpl_multi_plot(fn):
    """Decorate a plotting function to plot a grid of values.
    """

    @functools.wraps(fn)
    def multi_plotter(ds, *args, row=None, col=None, hspace=None, wspace=None,
                      tight_layout=True, coltitle=None, rowtitle=None,
                      **kwargs):

        if (row is None) and (col is None):
            return fn(ds, *args, **kwargs)

        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec

        # Set some global parameters
        p = fn(ds, *args, **kwargs, call=False)
        p.prepare_data_multi_grid()

        kwargs['vmin'] = kwargs.pop('vmin', p.vmin)
        kwargs['vmax'] = kwargs.pop('vmax', p.vmax)

        coltitle = col if coltitle is None else coltitle
        rowtitle = row if rowtitle is None else rowtitle

        # split the dataset into its respective rows and columns
        ds_r_c, nrows, ncols = calc_row_col_datasets(ds, row=row, col=col)

        figsize = kwargs.pop('figsize', (3 * ncols, 3 * nrows))
        return_fig = kwargs.pop('return_fig', PLOTTER_DEFAULTS['return_fig'])

        # generate a figure for all the plots to use
        p._fig = plt.figure(figsize=figsize, dpi=100,
                            constrained_layout=tight_layout)
        p._fig.set_constrained_layout_pads(hspace=hspace, wspace=wspace)
        # and a gridspec to position them
        gs = GridSpec(nrows=nrows, ncols=ncols, figure=p._fig,
                      hspace=hspace, wspace=wspace)

        # want to collect all entries for legend
        labels_handles = {}

        # range through rows and do subplots
        for i, ds_r in enumerate(ds_r_c):
            skws = {'legend': False, 'colorbar': False}

            # if not last row
            if i != nrows - 1:
                skws['xticklabels_hide'] = True
                skws['xtitle'] = ''

            # range through columns
            for j, sub_ds in enumerate(ds_r):

                if hspace == 0 and wspace == 0:
                    ticks_where = []
                    if j == 0:
                        ticks_where.append('left')
                    if i == 0:
                        ticks_where.append('top')
                    if j == ncols - 1:
                        ticks_where.append('right')
                    if i == nrows - 1:
                        ticks_where.append('bottom')
                    skws['ticks_where'] = ticks_where

                # if not first column
                if j != 0:
                    skws['yticklabels_hide'] = True
                    skws['ytitle'] = ''

                # label each column
                if (i == 0) and (col is not None):
                    col_val = prettify(ds[col].values[j])
                    skws['title'] = "{} = {}".format(coltitle, col_val)
                    fx = 'fontsize_xtitle'
                    skws['fontsize_title'] = kwargs.get(
                        fx, PLOTTER_DEFAULTS[fx])

                # label each row
                if (j == ncols - 1) and (row is not None):
                    # XXX: if number of cols==1 this hide yaxis - want both
                    row_val = prettify(ds[row].values[i])
                    skws['ytitle_right'] = True
                    skws['ytitle'] = "{} = {}".format(rowtitle, row_val)

                sP = fn(sub_ds, *args, add_to_fig=p._fig, call='both',
                        subplot=gs[i, j], **{**kwargs, **skws})

                try:
                    labels_handles.update(dict(zip(sP._legend_labels,
                                                   sP._legend_handles)))
                except AttributeError:
                    pass

        # make sure all have the same plot ranges
        xmins, xmaxs = zip(*(gax.get_xlim() for gax in p._fig.axes))
        ymins, ymaxs = zip(*(gax.get_ylim() for gax in p._fig.axes))
        xmin, xmax = min(xmins), max(xmaxs)
        ymin, ymax = min(ymins), max(ymaxs)
        for gax in p._fig.axes:
            gax.set_xlim(xmin, xmax)
            gax.set_ylim(ymin, ymax)

        # add global legend or colorbar
        p.plot_legend(grid=True, labels_handles=labels_handles)
        p.plot_colorbar(grid=True)

        if return_fig:
            plt.close(p._fig)
            return p._fig

    return multi_plotter


# --------------------------------------------------------------------------- #

class LinePlot(PlotterMatplotlib, AbstractLinePlot):
    """
    """

    def __init__(self, ds, x, y, z=None, *, y_err=None, x_err=None, **kwargs):
        super().__init__(ds, x, y, z=z, y_err=y_err, x_err=x_err, **kwargs)

    def plot_lines(self):
        """
        """
        for data in self._gen_xy():
            col = next(self._cols)

            line_opts = {
                'c': col,
                'lw': next(self._lws),
                'marker': next(self._mrkrs),
                'markersize': self._markersize,
                'markeredgecolor': col[:3] + (self.marker_alpha * col[3],),
                'markerfacecolor': col[:3] + (self.marker_alpha * col[3] / 2,),
                'label': next(self._zlbls),
                'zorder': next(self._zordrs),
                'linestyle': next(self._lines),
                'rasterized': self.rasterize,
            }

            if ('ye' in data) or ('xe' in data):
                self._axes.errorbar(data['x'], data['y'],
                                    yerr=data.get('ye', None),
                                    xerr=data.get('xe', None),
                                    ecolor=col,
                                    capsize=self.errorbar_capsize,
                                    capthick=self.errorbar_capthick,
                                    elinewidth=self.errorbar_linewidth,
                                    **line_opts)
            else:
                # add line to axes, with options cycled through
                self._axes.plot(data['x'], data['y'], **line_opts)

            self._legend_handles, self._legend_labels = \
                self._axes.get_legend_handles_labels()

    def __call__(self):
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_lines()
        self.plot_legend()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def lineplot(ds, x, y, z=None, y_err=None, x_err=None, **plot_opts):
    """From  ``ds`` plot lines of ``y`` as a function of ``x``, optionally for
    varying ``z``.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str or tuple[str]
        Variable(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return LinePlot(ds, x, y, z, y_err=y_err, x_err=x_err, **plot_opts)


class AutoLinePlot(LinePlot):
    def __init__(self, x, y_z, **lineplot_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, 'x', 'y', z='z', **lineplot_opts)


def auto_lineplot(x, y_z, **lineplot_opts):
    """Auto version of :func:`~xyzpy.lineplot` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoLinePlot(x, y_z, **lineplot_opts)()


# --------------------------------------------------------------------------- #

_SCATTER_ALT_DEFAULTS = (
    ('legend_handlelength', 0),
)


class Scatter(PlotterMatplotlib, AbstractScatter):

    def __init__(self, ds, x, y, z=None, **kwargs):
        # set some scatter specific options
        for k, default in _SCATTER_ALT_DEFAULTS:
            if k not in kwargs:
                kwargs[k] = default
        super().__init__(ds, x, y, z, **kwargs)

    def plot_scatter(self):
        """
        """
        self._legend_handles = []
        self._legend_labels = []

        for data in self._gen_xy():
            if 'c' in data:
                col = data['c']
            else:
                col = next(self._cols)

            scatter_opts = {
                'c': col,
                'marker': next(self._mrkrs),
                's': self._markersize,
                'alpha': self.marker_alpha,
                'label': next(self._zlbls),
                'zorder': next(self._zordrs),
                'rasterized': self.rasterize,
            }

            if 'c' in data:
                scatter_opts['cmap'] = self.cmap

            self._legend_handles.append(
                self._axes.scatter(data['x'], data['y'], **scatter_opts))
            self._legend_labels.append(
                scatter_opts['label'])

    def __call__(self):
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_scatter()
        self.plot_legend()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def scatter(ds, x, y, z=None, y_err=None, x_err=None, **plot_opts):
    """From  ``ds`` plot a scatter of ``y`` against ``x``, optionally for
    varying ``z``.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Quantity to plot along the x-axis.
    y : str or tuple[str]
        Quantity(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return Scatter(ds, x, y, z, y_err=y_err, x_err=x_err, **plot_opts)


class AutoScatter(Scatter):

    def __init__(self, x, y_z, **scatter_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, 'x', 'y', z='z', **scatter_opts)


def auto_scatter(x, y_z, **scatter_opts):
    """Auto version of :func:`~xyzpy.scatter` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoScatter(x, y_z, **scatter_opts)


# --------------------------------------------------------------------------- #

_HISTOGRAM_SPECIFIC_OPTIONS = {
    'stacked': False,
}

_HISTOGRAM_ALT_DEFAULTS = {
    'xtitle': 'x',
    'ytitle': 'f(x)',
}


class Histogram(PlotterMatplotlib, AbstractHistogram):

    def __init__(self, ds, x, z=None, **kwargs):

        # Set the alternative defaults
        for opt, default in _HISTOGRAM_ALT_DEFAULTS.items():
            if opt not in kwargs:
                kwargs[opt] = default

        # Set histogram specfic options
        for opt, default in _HISTOGRAM_SPECIFIC_OPTIONS.items():
            setattr(self, opt, kwargs.pop(opt, default))

        super().__init__(ds, x, None, z=z, **kwargs)

    def plot_histogram(self):
        from matplotlib.patches import Rectangle, Polygon

        def gen_ind_plots():
            for data in self._gen_xy():
                col = next(self._cols)

                edgecolor = col[:3] + (self.marker_alpha * col[3],)
                facecolor = col[:3] + (self.marker_alpha * col[3] / 4,)
                linewidth = next(self._lws)
                zorder = next(self._zordrs)
                label = next(self._zlbls)

                handle = Rectangle((0, 0), 1, 1, color=facecolor, ec=edgecolor)

                yield (data['x'], edgecolor, facecolor, linewidth, zorder,
                       label, handle)

        xs, ecs, fcs, lws, zds, lbs, hnds = zip(*gen_ind_plots())

        histogram_opts = {
            'label': lbs,
            'bins': self.bins,
            'density': True,
            'histtype': 'stepfilled',
            'fill': True,
            'stacked': self.stacked,
            'rasterized': self.rasterize,
        }

        _, _, patches = self._axes.hist(xs, **histogram_opts)

        # Need to set varying colors, linewidths etc seperately
        for patch, ec, fc, lw, zd in zip(patches, ecs, fcs, lws, zds):

            # patch is not iterable if only one set of data created
            if isinstance(patch, Polygon):
                patch = (patch,)

            for sub_patch in patch:
                sub_patch.set_edgecolor(ec)
                sub_patch.set_facecolor(fc)
                sub_patch.set_linewidth(lw)
                sub_patch.set_zorder(zd)

        # store handles for legend
        self._legend_handles, self._legend_labels = hnds, lbs

    def __call__(self):
        # Core preparation
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_histogram()
        self.plot_legend()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def histogram(ds, x, z=None, **plot_opts):
    """Dataset histogram.

    Parameters
    ----------
    ds : xarray.Dataset
        The dataset to plot.
    x : str, sequence of str
        The variable(s) to plot the probability density of. If sequence, plot a
        histogram of each instead of using a ``z`` coordinate.
    z : str, optional
        If given, range over this coordinate a plot a histogram for each.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return Histogram(ds, x, z=z, **plot_opts)


class AutoHistogram(Histogram):

    def __init__(self, x, **histogram_opts):
        ds = auto_xyz_ds(x)
        super().__init__(ds, 'x', **histogram_opts)


def auto_histogram(x, **histogram_opts):
    """Auto version of :func:`~xyzpy.histogram` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoHistogram(x, **histogram_opts)()


# --------------------------------------------------------------------------- #

_HEATMAP_ALT_DEFAULTS = (
    ('legend', False),
    ('colorbar', True),
    ('colormap', 'inferno'),
    ('method', 'pcolormesh'),
    ('gridlines', False),
    ('rasterize', True),
)


class HeatMap(PlotterMatplotlib, AbstractHeatMap):

    def __init__(self, ds, x, y, z, **kwargs):
        # set some heatmap specific options
        for k, default in _HEATMAP_ALT_DEFAULTS:
            if k not in kwargs:
                kwargs[k] = default
        super().__init__(ds, x, y, z, **kwargs)

    def plot_heatmap(self):
        """Plot the data as a heatmap.
        """
        self.calc_color_norm()
        self._heatmap = getattr(self._axes, self.method)(
            self._heatmap_x,
            self._heatmap_y,
            self._heatmap_var,
            norm=self._color_norm,
            cmap=xyz_colormaps(self.colormap),
            rasterized=self.rasterize)

    def __call__(self):
        # Core preparation
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_heatmap()
        self.plot_colorbar()
        self.set_panel_label()
        return self.show()


@mpl_multi_plot
@intercept_call_arg
def heatmap(ds, x, y, z, **kwargs):
    """From  ``ds`` plot variable ``z`` as a function of ``x`` and ``y`` using
    a 2D heatmap.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str
        Dimension to plot along the y-axis.
    z : str, optional
        Variable to plot as colormap.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return HeatMap(ds, x, y, z, **kwargs)


class AutoHeatMap(HeatMap):

    def __init__(self, x, **heatmap_opts):
        ds = auto_xyz_ds(x)
        super().__init__(ds, 'x', **heatmap_opts)


def auto_heatmap(x, **heatmap_opts):
    """Auto version of :func:`~xyzpy.heatmap` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoHeatMap(x, **heatmap_opts)()


# --------------- Miscellenous matplotlib plotting functions ---------------- #

def choose_squarest_grid(x):
    p = x ** 0.5
    if p.is_integer():
        m = n = int(p)
    else:
        m = int(round(p))
        p = int(p)
        n = p if m * p >= x else p + 1
    return m, n


def visualize_matrix(x, figsize=(4, 4),
                     colormap='Greys',
                     touching=False,
                     zlims=(None, None),
                     gridsize=None,
                     tri=None,
                     return_fig=True):
    """Plot the elements of one or more matrices.

    Parameters
    ----------
    x : array or iterable of arrays
        2d-matrix or matrices to plot.
    figsize : tuple
        Total size of plot.
    colormap : str
        Colormap to use to weight elements.
    touching : bool
        If plotting more than one matrix, whether the edges should touch.
    zlims:
        Scaling parameters for the element colorings, (i.e. if these are
        set then the weightings are not normalized).
    return_fig : bool
        Whether to return the figure created or just show it.
    """
    import matplotlib.pyplot as plt

    fig = plt.figure(figsize=figsize, dpi=100)
    if isinstance(x, np.ndarray):
        x = (x,)

    nx = len(x)
    if gridsize:
        m, n = gridsize
    else:
        m, n = choose_squarest_grid(nx)
    subplots = tuple((m, n, i) for i in range(1, nx + 1))

    for img, subplot in zip(x, subplots):

        if tri is not None:
            if tri not in {'upper', 'lower'}:
                raise ValueError("'tri' should be one of {'upper', 'lower}.")

            ma_fn = np.tril if tri == 'upper' else np.triu
            img = np.ma.array(img, mask=ma_fn(np.ones_like(img), k=-1))

        ax = fig.add_subplot(*subplot)
        ax.imshow(img, cmap=xyz_colormaps(colormap), interpolation='nearest',
                  vmin=zlims[0], vmax=zlims[1])
        # Hide the right and top spines
        ax.spines['right'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['top'].set_visible(False)
        # Only show ticks on the left and bottom spines
        ax.yaxis.set_visible(False)
        ax.xaxis.set_visible(False)
    plt.subplots_adjust(left=0, bottom=0, right=1, top=1,
                        wspace=-0.001 if touching else 0.05,
                        hspace=-0.001 if touching else 0.05)
    if return_fig:
        plt.close(fig)
        return fig
    else:
        plt.show()



"""
Helper functions for preparing data to be plotted.
"""
# TODO: x-err, upper and lower errors *************************************** #
# TODO: check shape and hint at which dimensions need to be reduced ********* #

import itertools
import functools
import numpy as np
import numpy.ma as ma
import xarray as xr
from .color import convert_colors, xyz_colormaps, get_default_sequential_cm
from .marker import _MARKERS, _SINGLE_LINE_MARKER

PLOTTER_DEFAULTS = {
    # Figure options
    'backend': 'MATPLOTLIB',
    'figsize': (7, 6),
    'axes_loc': None,
    'axes_rloc': None,
    'add_to_axes': None,
    'add_to_xaxes': None,
    'add_to_fig': None,
    'subplot': None,
    'fignum': 1,
    'return_fig': True,
    # Coloring options
    'colors': None,
    'colormap': None,
    'colormap_log': False,
    'colormap_reverse': False,
    # Colorbar options
    'colorbar': None,
    'ctitle': None,
    'vmin': None,
    'vmax': None,
    'colorbar_relative_position': None,
    'colorbar_where': None,
    'colorbar_opts': dict(),
    'method': None,
    'colorbar_color': "black",
    # 'z-axis' options: for legend or colorbar
    'ztitle': None,
    'zlabels': None,
    'zlims': (None, None),
    'zticks': None,
    'legend': None,
    'legend_loc': 'best',
    'legend_where': None,
    'legend_ncol': 1,
    'legend_bbox': None,
    'legend_marker_scale': None,
    'legend_label_spacing': None,
    'legend_column_spacing': 1,
    'legend_frame': False,
    'legend_handlelength': None,
    'legend_reverse': False,
    'legend_marker_alpha': None,
    # x-axis options
    'xtitle': None,
    'xtitle_pad': 5,
    'xlims': None,
    'xticks': None,
    'xtick_labels': None,
    'xticklabels_hide': False,
    'xlog': False,
    'bins': 30,
    # y-axis options
    'ytitle': None,
    'ytitle_pad': 5,
    'ytitle_right': False,
    'ylims': None,
    'yticks': None,
    'yticklabels_hide': False,
    'yticklabels_right': None,
    'ylog': False,
    # Titles and text
    'title': None,
    'panel_label': None,
    'panel_label_loc': (0.05, 0.93),
    'panel_label_color': 'black',
    # Styling options
    'markers': None,
    'markersize': None,
    'marker_alpha': 1.0,
    'lines': True,
    'line_styles': None,
    'line_widths': None,
    'errorbar_capthick': 1,
    'errorbar_capsize': 0,
    'errorbar_linewidth': 0.7,
    'zorders': None,
    'padding': None,
    'vlines': None,
    'hlines': None,
    'span_style': '--',
    'span_width': 1,
    'span_color': "0.5",
    'gridlines': True,
    'gridline_style': (1, 2),
    'ticks_where': ('bottom', 'left', 'top', 'right'),
    # Font options
    'font': ('CMU Serif', 'PT Serif', 'Liberation Serif', 'DejaVu Serif'),
    'fontsize_title': 18,
    'fontsize_ticks': 13,
    'fontsize_xtitle': 18,
    'fontsize_ytitle': 18,
    'fontsize_ztitle': 18,
    'fontsize_zlabels': 13,
    'fontsize_panel_label': 15,
    'math_serif': True,
    # backend options
    'rasterize': False,
    'webgl': False,
}

_PLOTTER_OPTS = list(PLOTTER_DEFAULTS.keys())


def check_excess_dims(ds, var, valid_dims, mode='lineplot'):

    if mode == 'lineplot':
        expanded_valid_dims = list(itertools.chain.from_iterable(
            ds[d].dims for d in valid_dims))

        excess_dims = {d for d, sz in ds[var].sizes .items()
                       if (sz > 1) and (d not in expanded_valid_dims)}
        if excess_dims:
            raise ValueError("Dataset has too many non-singlet dimensions "
                             "- try selection values for the following: "
                             "{}.".format(excess_dims))

    else:
        # TODO: for now, no checking if not lineplot
        pass


class Plotter:
    """
    """

    def __init__(self, ds, x, y, z=None, c=None, y_err=None, x_err=None,
                 **kwargs):
        """
        """
        if isinstance(ds, xr.DataArray):
            self._ds = ds.to_dataset()
        else:
            self._ds = ds

        self.x_coo = x
        self.y_coo = y
        self.z_coo = z
        self.c_coo = c
        self.y_err = y_err
        self.x_err = x_err

        # Figure options
        settings = {**PLOTTER_DEFAULTS, **kwargs}
        for opt in _PLOTTER_OPTS:
            setattr(self, opt, settings.pop(opt))

        # Parse unmatched keywords and error but suggest correct versions
        if len(settings) > 0:
            import difflib
            wrong_opts = list(settings.keys())
            right_opts = [difflib.get_close_matches(opt, _PLOTTER_OPTS, n=3)
                          for opt in wrong_opts]

            msg = ("Option(s) {} not valid.\n Did you mean: {}?"
                   .format(wrong_opts, right_opts))
            print(msg)
            raise ValueError(msg)

        if self.colors and self.c_coo:
            raise ValueError("Cannot specify explicit colors if ``c`` used.")

        # Internal state
        self._data_range_calculated = False
        self._c_cols = []

    # ----------------------------- properties ------------------------------ #

    def _get_ds(self):
        return self._ds

    def _set_ds(self, ds):
        self._ds = ds
        self.prepare_z_vals()
        self.prepare_z_labels()
        self.prepare_xy_vals_lineplot()
        self.calc_plot_range()
        self.update()

    def _del_ds(self):
        self._ds = None

    ds = property(_get_ds, _set_ds, _del_ds,
                  "The dataset used to plot graph.")

    # ------------------------------- methods ------------------------------- #

    def prepare_axes_labels(self):
        """Work out what the axes titles should be.
        """
        self._xtitle = self.x_coo if self.xtitle is None else self.xtitle
        if self.ytitle is None:
            if isinstance(self.y_coo, str):
                self._ytitle = self.y_coo
            else:
                self._ytitle = None
        else:
            self._ytitle = self.ytitle

        if self.c_coo is None:
            self._ctitle = self.z_coo if self.ztitle is None else self.ztitle
        else:
            self._ctitle = self.c_coo if self.ctitle is None else self.ctitle

    def prepare_z_vals(self, mode='lineplot', grid=False):
        """Work out what the 'z-coordinate', if any, should be.

        Parameters
        ----------
        mode : {'lineplot', 'scatter', 'histogram'}
        """
        # TODO: process multi-var errors at same time
        self._multi_var = False

        # Multiple sets of data parametized by z_coo
        if self.z_coo is not None:
            if not grid:
                check_excess_dims(self._ds, self.y_coo,
                                  (self.x_coo, self.z_coo), mode=mode)
            self._z_vals = self._ds[self.z_coo].values

        # Multiple data variables to plot -- just named in list
        elif isinstance(self.y_coo, (tuple, list)):
            for var in self.y_coo:
                if grid:
                    check_excess_dims(self._ds, var, (self.x_coo,), mode=mode)
            self._multi_var = True
            self._z_vals = self.y_coo

        # Multiple data variables to plot -- but for histogram
        elif isinstance(self.x_coo, (tuple, list)):
            self._multi_var = True
            self._z_vals = self.x_coo

        # Single data variable to plot
        else:
            if not grid:
                check_excess_dims(self._ds, self.y_coo,
                                  (self.x_coo,), mode=mode)
            self._z_vals = (None,)

    def prepare_z_labels(self):
        """Work out what the labels for the z-coordinate should be.
        """
        # Manually specified z-labels
        if self.zlabels is not None:
            self._zlbls = iter(self.zlabels)

        # Use z-data (for lineplot only) or multiple y names
        elif (self.z_coo is not None) or self._multi_var:
            self._zlbls = iter(str(z) for z in self._z_vals)

        # No z-labels
        else:
            self._zlbls = itertools.repeat(None)

    def calc_use_legend_or_colorbar(self):
        """Work out whether to use a legend.
        """
        def auto_legend():
            return (1 < len(self._z_vals) <= 10)

        # cspecified, lspecified, many points, colors, colors, c_coo
        if self.colorbar and (self.legend is None):
            self.legend = False if (self.c_coo is None) else auto_legend()
        if self.legend and (self.colorbar is None):
            self.colorbar = False if (self.c_coo is None) else True

        if self.legend is None and self.colorbar is None:
            self._use_legend = auto_legend()
            self._use_colorbar = (((not self._use_legend) and
                                   self.colors is True) or
                                  self.c_coo is not None)
        else:
            self._use_legend = self.legend
            self._use_colorbar = self.colorbar

    def prepare_xy_vals_lineplot(self, mode='lineplot'):
        """Select and flatten the data appropriately to iterate over.
        """

        def gen_xy():
            for i, z in enumerate(self._z_vals):
                das = {}
                data = {}

                # multiple data variables rather than z coordinate
                if self._multi_var:
                    das['x'] = self._ds[self.x_coo]
                    das['y'] = self._ds[z]

                    if (self.y_err is not None) or \
                       (self.x_err is not None) or \
                       (self.c_coo is not None):
                        raise ValueError('Multi-var errors/c not implemented.')

                # z-coordinate to iterate over
                elif z is not None:
                    try:
                        # try positional indexing first, as much faster
                        sub_ds = self._ds[{self.z_coo: i}]
                    except ValueError:
                        # but won't work e.g. on non-dimensions
                        sub_ds = self._ds.loc[{self.z_coo: z}]

                    das['x'] = sub_ds[self.x_coo]
                    das['y'] = sub_ds[self.y_coo]

                    if self.c_coo is not None:
                        if mode == 'lineplot':
                            self._c_cols.append(np.asscalar(
                                sub_ds[self.c_coo].values.flatten()))
                        elif mode == 'scatter':
                            das['c'] = sub_ds[self.c_coo]

                    if self.y_err is not None:
                        das['ye'] = sub_ds[self.y_err]

                    if self.x_err is not None:
                        das['xe'] = sub_ds[self.x_err]

                # nothing to iterate over
                else:
                    das['x'] = self._ds[self.x_coo]
                    das['y'] = self._ds[self.y_coo]

                    if self.c_coo is not None:
                        if mode == 'lineplot':
                            self._c_cols.append(np.asscalar(
                                self._ds[self.c_coo].values.flatten()))
                        elif mode == 'scatter':
                            das['c'] = self._ds[self.c_coo]

                    if self.y_err is not None:
                        das['ye'] = self._ds[self.y_err]

                    if self.x_err is not None:
                        das['xe'] = self._ds[self.x_err]

                for k, da in zip(das, xr.broadcast(*das.values())):
                    data[k] = da.values.flatten()

                # Trim out missing data
                not_null = np.isfinite(data['x'])
                not_null &= np.isfinite(data['y'])

                # TODO: if scatter, broadcast *then* ravel x, y, c?

                data['x'] = data['x'][not_null]
                data['y'] = data['y'][not_null]
                if 'c' in data:
                    data['c'] = data['c'][not_null]
                if 'ye' in data:
                    data['ye'] = data['ye'][not_null]
                if 'xe' in data:
                    data['xe'] = data['xe'][not_null]

                yield data

        self._gen_xy = gen_xy

    def prepare_x_vals_histogram(self):
        """
        """

        def gen_x():
            for z in self._z_vals:
                # multiple data variables rather than z coordinate
                if self._multi_var:
                    x = self._ds[z].values.flatten()

                # z-coordinate to iterate over
                elif z is not None:
                    sub_ds = self._ds.loc[{self.z_coo: z}]
                    x = sub_ds[self.x_coo].values.flatten()

                # nothing to iterate over
                else:
                    x = self._ds[self.x_coo].values.flatten()

                yield {'x': x[np.isfinite(x)]}

        self._gen_xy = gen_x

    def prepare_heatmap_data(self, grid=False):
        """Prepare the data to go into a heatmap.
        """
        # Can only show one dataset
        self._multi_var = False

        self._heatmap_x = self._ds[self.x_coo].values.flatten()
        self._heatmap_y = self._ds[self.y_coo].values.flatten()

        if grid:
            self._zmin = self._ds[self.z_coo].min()
            self._zmax = self._ds[self.z_coo].max()
            return

        self._heatmap_var = ma.masked_invalid(
            self._ds[self.z_coo]
                .squeeze()
                .transpose(self.y_coo, self.x_coo)
                .values)
        self._zmin = self._heatmap_var.min()
        self._zmax = self._heatmap_var.max()

    def calc_color_norm(self):
        import matplotlib as mpl

        self.cmap = xyz_colormaps(self.colormap, reverse=self.colormap_reverse)

        coo = self.z_coo if self.c_coo is None else self.c_coo

        if coo is None:
            return

        # check if real numeric type
        if self._ds[coo].dtype.kind in {'i', 'u', 'f'}:
            self._zmin = self.zlims[0]
            if self._zmin is None:
                self._zmin = self._ds[coo].values.min()
            self._zmax = self.zlims[1]
            if self._zmax is None:
                self._zmax = self._ds[coo].values.max()
        else:
            # no relative coloring possible e.g. for strings
            self._zmin, self._zmax = 0.0, 1.0

        if self.vmin is None:
            self.vmin = self._zmin
        if self.vmax is None:
            self.vmax = self._zmax

        self._color_norm = getattr(
            mpl.colors, "LogNorm" if self.colormap_log else "Normalize")(
                vmin=self.vmin, vmax=self.vmax)

        self.set_mappable()

    def calc_line_colors(self):
        """Helper function for calculating what the colormapped color of each
        line should be.
        """
        self.calc_color_norm()

        # set from 'adjacent' variable
        if self.c_coo is not None:
            rvals = (self._color_norm(z) for z in self._c_cols)
        # set from coordinate
        else:
            if np.isreal(self._z_vals[0]):
                rvals = (self._color_norm(z) for z in self._z_vals)
            else:
                # no relative coloring possible e.g. for strings
                rvals = np.linspace(0, 1, len(self._z_vals))

        # Map relative value to mpl color
        self._cols = (self.cmap(rval) for rval in rvals)
        # Convert colors to correct format
        self._cols = iter(convert_colors(self._cols, outformat=self.backend))

    def prepare_colors(self):
        """Prepare the colors for the lines, based on the z-coordinate.
        """
        # Automatic colors
        if (self.colors is True) or (self.c_coo is not None):
            self.calc_line_colors()
        # Manually specified colors
        elif self.colors:
            self._cols = itertools.cycle(
                convert_colors(self.colors, outformat=self.backend))
        # Use sequential
        else:
            self._cols = get_default_sequential_cm(self.backend)

    def prepare_markers(self):
        """Prepare the markers to be used for each line.
        """
        if self.markers is None:
            # If no lines, always draw markers
            if self.lines is False:
                self.markers = True
            # Else decide on how many points
            else:
                self.markers = len(self._ds[self.x_coo]) <= 51

        # Could add more logic based on number of xpoints?
        if self.markersize is None:
            self._markersize = 5
        else:
            self._markersize = self.markersize

        if self.markers:
            if len(self._z_vals) > 1:
                self._mrkrs = itertools.cycle(_MARKERS[self.backend])
            else:
                self._mrkrs = iter(_SINGLE_LINE_MARKER[self.backend])
        else:
            self._mrkrs = itertools.repeat(None)

    def prepare_line_styles(self):
        """Line widths and styles.
        """
        self._lines = (
            itertools.repeat(" ") if self.lines is False else
            itertools.repeat("solid") if self.line_styles is None else
            itertools.cycle(self.line_styles))

        # Set custom widths for each line
        if self.line_widths is not None:
            self._lws = itertools.cycle(self.line_widths)
        else:
            self._lws = itertools.cycle([1.3])

    def prepare_zorders(self):
        """What order lines appear over one another
        """
        if self.zorders is not None:
            self._zordrs = itertools.cycle(self.zorders)
        else:
            self._zordrs = itertools.cycle([3])

    def calc_data_range(self, force=False):
        """
        """
        if not self._data_range_calculated or force:
            # x data range
            if self.x_coo is None:
                self._data_xmin, self._data_xmax = None, None
            elif isinstance(self.x_coo, str):
                self._data_xmin = float(self._ds[self.x_coo].min())
                self._data_xmax = float(self._ds[self.x_coo].max())
            else:
                self._data_xmin = float(min(self._ds[x].min()
                                            for x in self.x_coo))
                self._data_xmax = float(max(self._ds[x].max()
                                            for x in self.x_coo))

            # y data range
            if self.y_coo is None:
                self._data_ymin, self._data_ymax = None, None
            elif isinstance(self.y_coo, str):
                self._data_ymin = float(self._ds[self.y_coo].min())
                self._data_ymax = float(self._ds[self.y_coo].max())
            else:
                self._data_ymin = float(min(self._ds[y].min()
                                            for y in self.y_coo))
                self._data_ymax = float(max(self._ds[y].max()
                                            for y in self.y_coo))

            self._data_range_calculated = True

    def calc_plot_range(self):
        """Logic for processing limits and padding into plot ranges.
        """
        # Leave as default
        if self.xlims is None and self.padding is None:
            self._xlims = None
        else:
            if self.xlims is not None:
                xmin, xmax = self.xlims
            else:
                self.calc_data_range()
                xmin, xmax = self._data_xmin, self._data_xmax

            if xmin is None and xmax is None:
                self._xlims = None

            # increase plot range if padding specified
            elif self.padding is not None:
                xrnge = xmax - xmin
                self._xlims = (xmin - self.padding * xrnge,
                               xmax + self.padding * xrnge)
            else:
                self._xlims = xmin, xmax

        if self.ylims is None and self.padding is None:
            # Leave as default
            self._ylims = None
        else:
            if self.ylims is not None:
                ymin, ymax = self.ylims
            else:
                self.calc_data_range()
                ymin, ymax = self._data_ymin, self._data_ymax

            if ymin is None and ymax is None:
                self._ylims = None

            # increase plot range if padding specified
            elif self.padding is not None:
                yrnge = ymax - ymin
                self._ylims = (ymin - self.padding * yrnge,
                               ymax + self.padding * yrnge)
            else:
                self._ylims = ymin, ymax


# Abstract Plotters - contain the core preparation required for each plot type

class AbstractLinePlot:

    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode='lineplot')
        self.prepare_z_labels()
        self.calc_use_legend_or_colorbar()
        self.prepare_xy_vals_lineplot(mode='lineplot')
        self.prepare_colors()
        self.prepare_markers()
        self.prepare_line_styles()
        self.prepare_zorders()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode='lineplot', grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()
        self.calc_data_range()


class AbstractScatter:

    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode='scatter')
        self.prepare_z_labels()
        self.calc_use_legend_or_colorbar()
        self.prepare_xy_vals_lineplot(mode='scatter')
        self.prepare_colors()
        self.prepare_markers()
        self.prepare_line_styles()
        self.prepare_zorders()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode='scatter', grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()
        self.calc_data_range()


class AbstractHistogram:

    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode='histogram')
        self.prepare_z_labels()
        self.calc_use_legend_or_colorbar()
        self.prepare_x_vals_histogram()
        self.prepare_colors()
        self.prepare_line_styles()
        self.prepare_zorders()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_z_vals(mode='histogram', grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()


class AbstractHeatMap:

    def prepare_data_single(self):
        self.prepare_axes_labels()
        self.prepare_heatmap_data()
        self.calc_use_legend_or_colorbar()
        self.calc_plot_range()

    def prepare_data_multi_grid(self):
        self.prepare_axes_labels()
        self.prepare_heatmap_data(grid=True)
        self.calc_use_legend_or_colorbar()
        self.calc_color_norm()
        self.calc_data_range()


# Helpers for grid plots

def calc_row_col_datasets(ds, row=None, col=None):
    """
    """
    if row is not None:
        rs = ds[row].values
        nr = len(rs)

    if col is not None:
        cs = ds[col].values
        nc = len(cs)

    if row is None:
        return [[ds.loc[{col: c}] for c in cs]], 1, nc

    if col is None:
        return [[ds.loc[{row: r}]] for r in rs], nr, 1

    return [[ds.loc[{row: r, col: c}] for c in cs] for r in rs], nr, nc


def intercept_call_arg(fn):

    @functools.wraps(fn)
    def wrapped_fn(*args, **kwargs):
        call = kwargs.pop('call', True)
        P = fn(*args, **kwargs)

        if call == 'both':
            P()
            return P
        elif call:
            return P()
        else:
            return P

    return wrapped_fn


def prettify(x):

    if np.issubdtype(type(x), np.floating):
        x = "{0:0.4f}".format(x).rstrip('0')
        if x[-1] == '.':
            x += "0"

    return x



"""Functions for systematically evaluating a function over all combinations.
"""
import functools
from concurrent import futures as cf
import numpy as np
import xarray as xr

from ..utils import (
    unzip,
    flatten,
    prod,
    progbar,
)
from .prepare import (
    _parse_var_names,
    _parse_var_dims,
    _parse_constants,
    _parse_resources,
    _parse_var_coords,
    _parse_combos,
    _parse_combo_results,
)


def default_submitter(pool, fn, *args, **kwds):
    """Default method for submitting to a pool.
    """
    try:
        future = pool.submit(fn, *args, **kwds)
    except AttributeError:
        future = pool.apply_async(fn, *args, **kwds)
    return future


def nested_submit(fn, combos, kwds, pool=None, submitter=default_submitter):
    """Recursively submit jobs directly, as delayed objects or to a pool.

    Parameters
    ----------
    fn : callable
        Function to submit jobs to.
    combos : tuple mapping individual fn arguments to sequence of values
        Mapping of each argument and all its possible values.
    kwds : dict
        Constant keyword arguments not to iterate over.
    pool : Executor pool
        Pool-executor-like class implementing a submit method.

    Returns
    -------
    results : nested tuple
        Nested tuples of results.
    """
    arg, inputs = combos[0]

    # have reached most nested level?
    if len(combos) == 1:
        if pool:
            return tuple(submitter(pool, fn, **kwds, **{arg: x})
                         for x in inputs)
        else:
            return tuple(fn(**kwds, **{arg: x}) for x in inputs)
    # else recurse through current level
    else:
        return tuple(nested_submit(fn, combos[1:], {**kwds, arg: x}, pool=pool,
                                   submitter=submitter) for x in inputs)


def default_getter(pbar=None):
    """Generate the default function to get a result from a future, updating
    the progress bar ``pbar`` in the process.
    """
    if pbar:
        def getter(future):
            try:
                res = future.result()
            except AttributeError:
                res = future.get()
            pbar.update()
            return res
    else:
        def getter(future):
            try:
                res = future.result()
            except AttributeError:
                res = future.get()
            return res
    return getter


def nested_get(futures, ndim, getter):
    """Recusively get results from nested futures.
    """
    return (tuple(getter(fut) for fut in futures) if ndim == 1 else
            tuple(nested_get(fut, ndim - 1, getter) for fut in futures))


def _combo_runner_pool(fn, combos, constants, n, ndim, pool, verbosity=1):
    """Submit and retrieve combos from a generic pool.
    """
    with progbar(total=n, disable=verbosity <= 0) as pbar:

        if verbosity >= 2:
            pbar.set_description("Processing with pool")

        futures = nested_submit(fn, combos, constants, pool=pool)
        getter = default_getter(pbar)
        return nested_get(futures, ndim, getter)


def _combo_runner_parallel(fn, combos, constants, n, ndim,
                           num_workers, verbosity=1):
    """Submit and retrieve combos from a ProcessPoolExecutor.
    """
    with cf.ProcessPoolExecutor(max_workers=num_workers) as pool:

        with progbar(total=n, disable=verbosity <= 0) as pbar:

            if verbosity >= 2:
                pbar.set_description(
                    "Processing with {} workers".format(pool._max_workers)
                )

            futures = nested_submit(fn, combos, constants, pool=pool)
            for f in cf.as_completed(flatten(futures, ndim)):
                pbar.update()
            return nested_get(futures, ndim, default_getter())


def update_upon_eval(fn, pbar, verbosity=1):
    """Decorate `fn` such that every time it is called, `pbar` is updated
    """

    @functools.wraps(fn)
    def new_fn(**kwargs):
        if verbosity >= 2:
            pbar.set_description(str(kwargs))
        result = fn(**kwargs)
        pbar.update()
        return result

    return new_fn


def _combo_runner_sequential(fn, combos, constants, n, ndim, verbosity=1):
    """Run combos in a sequential manner.
    """
    with progbar(total=n, disable=verbosity <= 0) as pbar:

        # Wrap the function such that the progbar is updated upon each call
        fn = update_upon_eval(fn, pbar, verbosity=verbosity)
        return nested_submit(fn, combos, constants)


def _combo_runner(fn, combos, constants, split=False,
                  parallel=False, num_workers=None, pool=None,
                  verbosity=1):
    """Core combo runner, i.e. no parsing of arguments.
    """
    n = prod(len(x) for _, x in combos)
    ndim = len(combos)

    kws = {'fn': fn, 'combos': combos, 'constants': constants, 'n': n,
           'ndim': ndim, 'verbosity': verbosity}

    # Custom pool supplied
    if pool is not None:
        results = _combo_runner_pool(pool=pool, **kws)

    # Else for parallel, by default use a process pool exceutor
    elif parallel or num_workers:
        results = _combo_runner_parallel(num_workers=num_workers, **kws)

    # Evaluate combos sequentially
    else:
        results = _combo_runner_sequential(**kws)

    return tuple(unzip(results, ndim)) if split else results


def combo_runner(fn, combos, *, constants=None, split=False,
                 parallel=False, pool=None, num_workers=None,
                 verbosity=1):
    """Take a function fn and analyse it over all combinations of named
    variables' values, optionally showing progress and in parallel.

    Parameters
    ----------
    fn : callable
        Function to analyse.
    combos : mapping of individual fn arguments to sequence of values
        All combinations of each argument will be calculated. Each
        argument range thus gets a dimension in the output array(s).
    constants : dict, optional
        List of tuples/dict of *constant* fn argument mappings.
    split : bool, optional
        Whether to split (unzip) into multiple output arrays or not.
    parallel : bool, optional
        Process combos in parallel, default number of workers picked.
    pool : executor-like pool, optional
        Submit all combos to this pool. Must have ``submit`` or ``apply_async``
        methods.
    num_workers : int, optional
        Explicitly choose how many workers to use, None for automatic.
    verbosity : {0, 1, 2}, optional
        How much information to display:

        - 0: nothing,
        - 1: just progress,
        - 2: all information.

    Returns
    -------
    data : nested tuple
        Nested tuple containing all combinations of running ``fn``.
    """
    # Prepare combos
    combos = _parse_combos(combos)
    constants = _parse_constants(constants)

    # Submit to core combo runner
    return _combo_runner(fn, combos, constants=constants, split=split,
                         parallel=parallel, pool=pool, num_workers=num_workers,
                         verbosity=verbosity)


def multi_concat(results, dims):
    """Concatenate a nested list of xarray objects along several dimensions.
    """
    if len(dims) == 1:
        return xr.concat(results, dim=dims[0])
    else:
        return xr.concat([multi_concat(sub_results, dims[1:])
                          for sub_results in results], dim=dims[0])


def get_ndim_first(x, ndim):
    """Return the first element from the ndim-nested list x.
    """
    return (x if ndim == 0 else
            get_ndim_first(x[0], ndim - 1))


def _combos_to_ds(results, combos, var_names, var_dims, var_coords,
                  constants=None, attrs=None):
    """Convert the output of combo_runner into a `xarray.Dataset`.
    """
    fn_args = tuple(x for x, _ in combos)
    results = _parse_combo_results(results, var_names)

    # Check if the results are an array of xarray objects
    xobj_results = isinstance(get_ndim_first(results, len(fn_args) + 1),
                              (xr.Dataset, xr.DataArray))

    if xobj_results:
        # concat them all together, no var_names needed
        ds = multi_concat(results[0], fn_args)
        # Set dataset coordinates
        for fn_arg, vals in combos:
            ds[fn_arg] = vals
    else:
        # create a new dataset using the given arrays and var_names
        ds = xr.Dataset(
            coords={
                **dict(combos),
                **dict(var_coords)
            },
            data_vars={
                name: (fn_args + var_dims[name], np.asarray(data))
                for data, name in zip(results, var_names)
            })

    if attrs:
        ds.attrs = attrs

    # Add constants to attrs, but filter out those which should be coords
    if constants:
        for k, v in constants.items():
            if k in ds.dims:
                ds.coords[k] = v
            else:
                ds.attrs[k] = v
    return ds


def combo_runner_to_ds(fn, combos, var_names, *,
                       var_dims=None,
                       var_coords=None,
                       constants=None,
                       resources=None,
                       attrs=None,
                       parse=True,
                       **combo_runner_settings):
    """Evaluate a function over all combinations and output to a Dataset.

    Parameters
    ----------
    fn : callable
        Function to evaluate.
    combos : mapping
        Mapping of each individual function argument to sequence of values.
    var_names : str, sequence of strings, or None
        Variable name(s) of the output(s) of `fn`, set to None if
        fn outputs data already labelled in a Dataset or DataArray.
    var_dims : sequence of either strings or string sequences, optional
        'Internal' names of dimensions for each variable, the values for
        each dimension should be contained as a mapping in either
        `var_coords` (not needed by `fn`) or `constants` (needed by `fn`).
    var_coords : mapping, optional
        Mapping of extra coords the output variables may depend on.
    constants : mapping, optional
        Arguments to `fn` which are not iterated over, these will be
        recorded either as attributes or coordinates if they are named
        in `var_dims`.
    resources : mapping, optional
        Like `constants` but they will not be recorded.
    attrs : mapping, optional
        Any extra attributes to store.
    combo_runner_settings
        Arguments supplied to :func:`~xyzpy.combo_runner`.

    Returns
    -------
    ds : xarray.Dataset
        Multidimensional labelled dataset contatining all the results.
    """
    if parse:
        combos = _parse_combos(combos)
        var_names = _parse_var_names(var_names)
        var_dims = _parse_var_dims(var_dims, var_names=var_names)
        var_coords = _parse_var_coords(var_coords)
        constants = _parse_constants(constants)
        resources = _parse_resources(resources)

    # Generate data for all combos
    results = _combo_runner(fn, combos, constants={**resources, **constants},
                            split=len(var_names) > 1,
                            **combo_runner_settings)
    # Convert to dataset
    ds = _combos_to_ds(results, combos,
                       var_names=var_names,
                       var_dims=var_dims,
                       var_coords=var_coords,
                       constants=constants,
                       attrs=attrs)
    return ds



"""Utility functions.
"""
import functools
import operator
import itertools
import tqdm
import time


def prod(it):
    """Product of an iterable.
    """
    return functools.reduce(operator.mul, it)


def unzip(its, zip_level=1):
    """Split a nested iterable at a specified level, i.e. in numpy language
    transpose the specified 'axis' to be the first.

    Parameters
    ----------
    its: iterable (of iterables (of iterables ...))
        'n-dimensional' iterable to split
    zip_level: int
        level at which to split the iterable, default of 1 replicates
        ``zip(*its)`` behaviour.

    Example
    -------
    >>> x = [[(1, True), (2, False), (3, True)],
             [(7, True), (8, False), (9, True)]]
    >>> nums, bools = unzip(x, 2)
    >>> nums
    ((1, 2, 3), (7, 8, 9))
    >>> bools
    ((True, False, True), (True, False, True))

    """
    def _unzipper(its, zip_level):
        if zip_level > 1:
            return (zip(*_unzipper(it, zip_level - 1)) for it in its)
        else:
            return its

    return zip(*_unzipper(its, zip_level)) if zip_level else its


def flatten(its, n):
    """Take the n-dimensional nested iterable its and flatten it.

    Parameters
    ----------
        its : nested iterable
        n : number of dimensions

    Returns
    -------
        flattened iterable of all items
    """
    if n > 1:
        return itertools.chain(*(flatten(it, n - 1) for it in its))
    else:
        return its


def _get_fn_name(fn):
    """Try to inspect a function's name, taking into account several common
    non-standard types of function: dask, functools.partial ...
    """
    if hasattr(fn, "__name__"):
        return fn.__name__
    # try dask delayed function with key
    elif hasattr(fn, "key"):
        return fn.key.partition('-')[0]
    # try functools.partial function syntax
    elif hasattr(fn, "func"):
        return fn.func.__name__
    else:
        raise ValueError("Could not extract function name from {}".format(fn))


def progbar(it=None, nb=False, **kwargs):
    """Turn any iterable into a progress bar, with notebook option

    Parameters
    ----------
        it: iterable
            Iterable to wrap with progress bar
        nb: bool
            Whether  to display the notebook progress bar
        **kwargs: dict-like
            additional options to send to tqdm
    """
    defaults = {'ascii': True, 'smoothing': 0.0}
    # Overide defaults with custom kwargs
    settings = {**defaults, **kwargs}
    if nb:  # pragma: no cover
        return tqdm.tqdm_notebook(it, **settings)
    return tqdm.tqdm(it, **settings)


class Timer:
    """A simple context manager class for timing blocks.
    """

    def __enter__(self):
        self.start = time.time()
        return self

    def __exit__(self, *args):
        self.end = time.time()
        self.interval = self.end - self.start



"""Processing and signal analysis.
"""

import functools

import numpy as np
from scipy import interpolate, signal
import xarray as xr
from xarray.core.computation import apply_ufunc
from scipy.interpolate import LSQUnivariateSpline

try:
    from numba import njit, guvectorize, double, int_
except ImportError:

    no_numba_msg = "This function requires ``numba`` to be installed."

    def njit(*args, **kwargs):
        def raise_error_fn(*args, **kwargs):
            raise ImportError(no_numba_msg)
        return raise_error_fn

    def guvectorize(*args, **kwargs):
        raise ImportError(no_numba_msg)

    # define these just to avoid NameError
    double, int_ = (), ()


_NUMBA_CACHE_DEFAULT = False


class LazyCompile(object):
    """Class that 'compiles' a function only when called.
    """

    def __init__(self, fn, compiler, compiler_args, compiler_kwargs):
        self.fn = fn
        self.compiler = compiler
        self.compiled = False
        self.compiler_args = compiler_args
        self.compiler_kwargs = compiler_kwargs

    def compile_fn(self):
        self._fn = self.compiler(*self.compiler_args,
                                 **self.compiler_kwargs)(self.fn)
        self.compiled = True

    def __call__(self, *args, **kwargs):
        if not self.compiled:
            self.compile_fn()
        return self._fn(*args, **kwargs)


def lazy_guvectorize(*gufunc_args, **gufunc_kwargs):
    """Function to wrap LazyCompile around functions.
    """
    def actual_fn_wrapper(fn):
        lazy_fn = LazyCompile(fn, guvectorize, gufunc_args, gufunc_kwargs)

        @functools.wraps(fn)
        def cached_function(*args, **kwargs):
            return lazy_fn(*args, **kwargs)

        return cached_function

    return actual_fn_wrapper


@njit
def preprocess_nan_func(x, y, out):  # pragma: no cover
    """Pre-process data for a 1d function that doesn't accept nan-values.
    """
    # strip out nan
    mask = np.isfinite(x) & np.isfinite(y)
    num_nan = np.sum(~mask)

    if num_nan == x.size:
        out[:] = np.nan
        return
    elif num_nan != 0:
        x = x[mask]
        y = y[mask]

    return x, y, num_nan, mask


@njit
def postprocess_nan_func(yf, num_nan, mask, out):  # pragma: no cover
    """Post-process data for a 1d function that doesn't accept nan-values.
    """
    if num_nan != 0:
        out[~mask] = np.nan
        out[mask] = yf
    else:
        out[:] = yf


def preprocess_interp1d_nan_func(x, y, out):
    """Pre-process data for a 1d function that doesn't accept nan-values and
    needs evenly spaced data.
    """
    # strip out nan
    mask = np.isfinite(x) & np.isfinite(y)
    num_nan = np.sum(~mask)

    if num_nan == x.size:
        out[:] = np.nan
        return
    elif num_nan != 0:
        x = x[mask]
        y = y[mask]

    # interpolate to evenly spaced grid
    x_even = np.linspace(x.min(), x.max(), x.size)
    y_even = np.interp(x_even, x, y)
    return x, y, x_even, y_even, num_nan, mask


def postprocess_interp1d_nan_func(x, x_even, yf_even, num_nan, mask, out):
    """Post-process data for a 1d function that doesn't accept nan-values and
    needs evenly spaced data.
    """
    # re-interpolate to original spacing
    yf = np.interp(x, x_even, yf_even)

    if num_nan != 0:
        out[~mask] = np.nan
        out[mask] = yf
    else:
        out[:] = yf


# --------------------------------------------------------------------------- #
#                    fornberg's finite difference algortihm                   #
# --------------------------------------------------------------------------- #

@njit
def diff_fornberg(fx, x, z, order):  # pragma: no cover
    """Fornberg finite difference method for single point `z`.
    """
    c1 = 1.0
    c4 = x[0] - z
    n = len(x) - 1
    c = np.zeros((len(x), order + 1))
    c[0, 0] = 1.0

    for i in range(1, n + 1):
        mn = min(i, order)
        c2 = 1.0
        c5 = c4
        c4 = x[i] - z

        for j in range(0, i):
            c3 = x[i] - x[j]
            c2 *= c3

            if j == i - 1:

                for k in range(mn, 0, -1):
                    c[i, k] = c1 * (k * c[i - 1, k - 1] -
                                    c5 * c[i - 1, k]) / c2
                c[i, 0] = -c1 * c5 * c[i - 1, 0] / c2

            for k in range(mn, 0, -1):
                c[j, k] = (c4 * c[j, k] - k * c[j, k - 1]) / c3

            c[j, 0] = c4 * c[j, 0] / c3

        c1 = c2

    return np.dot(c[:, order], fx)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:], double[:]),
], '(n),(n),(m),(),()->(m)', cache=_NUMBA_CACHE_DEFAULT, nopython=True)
def finite_diff_array(fx, x, ix, order, window, out=None):  # pragma: no cover
    """Fornberg finite difference method for array of points `ix`.
    """
    fx = fx.astype(np.float64)

    w = window[0]
    if w < 0:  # use whole window
        for i, z in enumerate(ix):
            out[i] = diff_fornberg(fx, x, z, order[0])
    else:
        forward_limit = (x[0] + w / 2)
        foward_win = x[0] + w
        backward_limit = (x[-1] - w / 2)
        backward_win = x[-1] - w

        for i, z in enumerate(ix):
            if z < forward_limit:  # use forward diff
                bm = np.less(x, foward_win)
            elif z > backward_limit:  # backward diff
                bm = np.greater(x, backward_win)
            else:  # central diff
                bm = np.less(np.abs(x - z), w)
            wx = x[bm]
            wfx = fx[bm]
            out[i] = diff_fornberg(wfx, wx, z, order[0])


def _diff_fornberg_broadcast(x, fx, ix, order, mode='points',
                             window=5, axis=-1):
    """Find (d^k fx)/(dx^k) at points ix, using a windowed finite difference.
    This is only appropirate for very nicely sampled/analytic data.

    Uses algorithm found in:
        Calculation of Weights in Finite Difference Formulas
        Bengt Fornberg
        SIAM Rev., 40(3), 685–691
        http://dx.doi.org/10.1137/S0036144596322507

    Parameters
    ----------
    fx : array
        Function values at grid values.
    x : array
        Grid values, same legnth as `fx`, assumed sorted.
    ix : array or int
        If array, values at which to evalute finite difference else number
        of points to linearly space in the range and use.
    order : int
        Order of derivate, 0 yields an interpolation.
    mode : {'points', 'relative', 'absolute'}, optional
        Used in conjuction with window.
        If 'points', the window size is set such that the average number
        of points in each window is given b `window`.
        If 'relative', the window size is set such that its size relative
        to the total range is given by `window`.
        If 'absolute', the window size is geven explicitly by `window`.
    window : int or float, optional
        Depends on `mode`,
            - 'points', target number of points to use for each window.
            - 'relative' relative size of window compared to full range.
            - 'absolute' The absolute window size.
    return_func : bool, optional
        If True, return the single argument function wfdiff(fx).

    Returns
    -------
    ifx : array
        Interpolated kth-derivative of data.
    """
    if axis != -1:
        fx = fx.swapaxes(axis, -1)

    if mode in {'p', 'pts', 'points'}:
        abs_win = (x[-1] - x[0]) * window / len(x)
    elif mode in {'r', 'rel', 'relative'}:
        abs_win = (x[-1] - x[0]) * window
    elif mode in {'a', 'abs', 'absolute'}:
        abs_win = float(window)
    else:
        raise ValueError("mode: {} not valid".format(mode))

    if ix is None:
        ix = x.astype(float)
    elif isinstance(ix, int):
        ix = np.linspace(x[0], x[-1], ix)
    elif ix.dtype != float:
        ix = ix.astype(float)

    return finite_diff_array(fx, x, ix, order, abs_win)


def xr_diff_fornberg(obj, dim, ix=100, order=1, mode='points', window=5):
    """Find ``(d^k fx)/(dx^k)`` at points ``ix``, using a windowed finite
    difference. This is only appropirate for very nicely sampled/analytic data.

    Uses algorithm found in:
        Calculation of Weights in Finite Difference Formulas
        Bengt Fornberg
        SIAM Rev., 40(3), 685–691
        http://dx.doi.org/10.1137/S0036144596322507

    Parameters
    ----------
    xobj : xarray.DataArray or xarray.Dataset
        Object to find windowed finite difference for.
    dim : str
        Dimension to find windowed finite difference along.
    ix : array or int
        If array, values at which to evalute finite difference else number
        of points to linearly space in the range and use.
    order : int
        Order of derivate, 0 yields an interpolation.
    mode : {'points', 'relative', 'absolute'}
        Used in conjuction with window.
        If 'points', the window size is set such that the average number
        of points in each window is given by `window`.
        If 'relative', the window size is set such that its size relative
        to the total range is given by `window`.
        If 'absolute', the window size is geven explicitly by `window`.
    window : int or float
        Depends on ``mode``:

        - 'points', target number of points to use for each window.
        - 'relative' relative size of window compared to full range.
        - 'absolute' The absolute window size.

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset
        Object now with windowed finite difference along `dim`.
    """
    input_core_dims = [(dim,), (dim,)]
    args = (obj[dim], obj)

    if ix is None:
        kwargs = {'ix': ix, 'axis': -1, 'order': order,
                  'mode': mode, 'window': window}

        output_core_dims = [(dim,)]
        return apply_ufunc(_diff_fornberg_broadcast, *args, kwargs=kwargs,
                           input_core_dims=input_core_dims,
                           output_core_dims=output_core_dims)

    if isinstance(ix, int):
        ix = np.linspace(float(obj[dim].min()), float(obj[dim].max()), ix)

    kwargs = {'ix': ix, 'axis': -1, 'order': order,
              'mode': mode, 'window': window}
    output_core_dims = [('__temp_dim__',)]

    result = apply_ufunc(_diff_fornberg_broadcast, *args, kwargs=kwargs,
                         input_core_dims=input_core_dims,
                         output_core_dims=output_core_dims)
    result['__temp_dim__'] = ix
    return result.rename({'__temp_dim__': dim})


# --------------------------------------------------------------------------- #
#                      Unevenly spaced finite difference                      #
# --------------------------------------------------------------------------- #

@lazy_guvectorize([
    (int_[:], int_[:], double[:]),
    (double[:], int_[:], double[:]),
    (int_[:], double[:], double[:]),
    (double[:], double[:], double[:]),
], "(n),(n)->(n)", cache=_NUMBA_CACHE_DEFAULT, nopython=True)
def diff_u(fx, x, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, fx, out)
    if xynm is None:
        return
    x, fx, num_nan, mask = xynm
    yf = np.empty_like(fx)

    n = len(x)

    # Forward difference for first point
    h1 = x[1] - x[0]
    h2 = x[2] - x[1]
    yf[0] = (- fx[0] * (2 * h1 + h2) / (h1 * (h1 + h2)) +
             fx[1] * (h1 + h2) / (h1 * h2) -
             fx[2] * h1 / ((h1 + h2) * h2))

    # Central difference for middle points
    for i in range(1, n - 1):
        h1 = x[i] - x[i - 1]
        h2 = x[i + 1] - x[i]
        yf[i] = (- fx[i - 1] * h2 / (h1 * (h1 + h2)) -
                 fx[i] * (h1 - h2) / (h1 * h2) +
                 fx[i + 1] * h1 / (h2 * (h1 + h2)))

    # Backwards difference for last point
    h1 = x[n - 2] - x[n - 3]
    h2 = x[n - 1] - x[n - 2]
    yf[n - 1] = (fx[n - 3] * h2 / (h1 * (h1 + h2)) -
                 fx[n - 2] * (h1 + h2) / (h1 * h2) +
                 fx[n - 1] * (h1 + 2 * h2) / (h2 * (h1 + h2)))

    postprocess_nan_func(yf, num_nan, mask, out)


def _broadcast_diff_u(x, fx, axis=-1):
    if axis != -1:
        fx = fx.swapaxes(axis, -1)
    return diff_u(fx, x)


def xr_diff_u(obj, dim):
    """Uneven-third-order finite difference derivative [1].

    [1] Singh, Ashok K., and B. S. Bhadauria. "Finite difference formulae for
    unequal sub-intervals using lagrange’s interpolation formula."
    International Journal of Mathematics and Analysis 3.17 (2009): 815-827.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        The object to differentiate.
    dim : str
        The dimension to differentiate along.

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset
    """
    kwargs = {'axis': -1}
    input_core_dims = [(dim,), (dim,)]
    output_core_dims = [(dim,)]
    args = (obj[dim], obj)
    return apply_ufunc(_broadcast_diff_u, *args,
                       input_core_dims=input_core_dims,
                       output_core_dims=output_core_dims,
                       kwargs=kwargs)


@lazy_guvectorize([
    (int_[:], int_[:], double[:]),
    (double[:], int_[:], double[:]),
    (int_[:], double[:], double[:]),
    (double[:], double[:], double[:]),
], "(n),(n)->(n)", cache=_NUMBA_CACHE_DEFAULT, nopython=True)
def diff_u_err(efx, x, out=None):  # pragma: no cover
    """Propagate uncertainties using uneven finite difference formula.
    """
    xynm = preprocess_nan_func(x, efx, out)
    if xynm is None:
        return
    x, efx, num_nan, mask = xynm
    yf = np.empty_like(efx)

    n = len(x)
    yf = np.empty_like(efx)

    # Forward difference for first point
    h1 = x[1] - x[0]
    h2 = x[2] - x[1]
    yf[0] = ((efx[0] * (2 * h1 + h2) / (h1 * (h1 + h2)))**2 +
             (efx[1] * (h1 + h2) / (h1 * h2))**2 +
             (efx[2] * h1 / ((h1 + h2) * h2))**2)**0.5

    # Central difference for middle points
    for i in range(1, n - 1):
        h1 = x[i] - x[i - 1]
        h2 = x[i + 1] - x[i]
        yf[i] = ((efx[i - 1] * h2 / (h1 * (h1 + h2)))**2 +
                 (efx[i] * (h1 - h2) / (h1 * h2))**2 +
                 (efx[i + 1] * h1 / (h2 * (h1 + h2)))**2)**0.5

    # Backwards difference for last point
    h1 = x[n - 2] - x[n - 3]
    h2 = x[n - 1] - x[n - 2]
    yf[n - 1] = ((efx[n - 3] * h2 / (h1 * (h1 + h2)))**2 +
                 (efx[n - 2] * (h1 + h2) / (h1 * h2))**2 +
                 (efx[n - 1] * (h1 + 2 * h2) / (h2 * (h1 + h2)))**2)**0.5

    postprocess_nan_func(yf, num_nan, mask, out)


def _broadcast_diff_u_err(x, fx, axis=-1):
    if axis != -1:
        fx = fx.swapaxes(axis, -1)
    return diff_u_err(fx, x)


def xr_diff_u_err(obj, dim):
    """Propagate error through uneven-third-order finite difference derivative.
    If you have calculated a derivative already using ``xr_diff_u``, and you
    have data about the uncertainty on the original data, this function
    propagates that error through to be an error on the derivative.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        The object to differentiate.
    dim : str
        The dimension to differentiate along.

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset
    """
    kwargs = {'axis': -1}
    input_core_dims = [(dim,), (dim,)]
    output_core_dims = [(dim,)]
    args = (obj[dim], obj)
    return apply_ufunc(_broadcast_diff_u_err, *args,
                       input_core_dims=input_core_dims,
                       output_core_dims=output_core_dims,
                       kwargs=kwargs)


# --------------------------------------------------------------------------- #
#                                interpolation                                #
# --------------------------------------------------------------------------- #

_INTERP_INT2STR = {
    0: 'zero',
    1: 'slinear',
    2: 'quadratic',
    3: 'cubic',
}


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:]),
    (double[:], int_[:], int_[:], double[:]),
    (int_[:], double[:], int_[:], double[:]),
    (double[:], double[:], int_[:], double[:]),
], '(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_interp(x, y, order, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, num_nan, mask = xynm

    # interpolating function
    ifn = interpolate.interp1d(x, y, kind=_INTERP_INT2STR[order[0]],
                               bounds_error=False)
    yf = ifn(x)
    postprocess_nan_func(yf, num_nan, mask, out)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:])
], '(n),(n),(m),()->(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_interp_upscale(x, y, ix, order, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, _, _ = xynm

    # interpolating function
    ifn = interpolate.interp1d(x, y, kind=_INTERP_INT2STR[order[0]],
                               bounds_error=False)

    # no need to nan_func post process and upscaling into exact right size
    out[:] = ifn(ix)


def _broadcast_interp(x, y, ix=100, order=3, axis=-1):
    if axis != -1:
        y = y.swapaxes(axis, -1)

    if ix is None:
        return _gufunc_interp(x, y, order)

    if isinstance(ix, int):
        # automatic upscale
        ix = np.linspace(x.min(), x.max(), ix)

    return _gufunc_interp_upscale(x, y, ix, order)


def xr_interp(obj, dim, ix=100, order=3):
    """Interpolate along axis ``dim`` using :func:`scipy.interpolate.interp1d`.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        The object to interpolate.
    dim : str
        The axis to interpolate along.
    ix : int or array
        If int, interpolate to this many points spaced evenly along the range
        of the original data. If array, interpolate to those points directly.
    order : int
        Supplied to :func:`scipy.interpolate.interp1d` as the order of
        interpolation.

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset

    See Also
    --------
    xr_pchip
    """

    input_core_dims = [(dim,), (dim,)]
    args = (obj[dim], obj)
    kwargs = {'ix': ix, 'axis': -1, 'order': order}

    if ix is None:
        output_core_dims = [(dim,)]
        return apply_ufunc(_broadcast_interp, *args, kwargs=kwargs,
                           input_core_dims=input_core_dims,
                           output_core_dims=output_core_dims)

    if isinstance(ix, int):
        ix = np.linspace(float(obj[dim].min()), float(obj[dim].max()), ix)

    kwargs['ix'] = ix
    output_core_dims = [('__temp_dim__',)]

    result = apply_ufunc(_broadcast_interp, *args, kwargs=kwargs,
                         input_core_dims=input_core_dims,
                         output_core_dims=output_core_dims)
    result['__temp_dim__'] = ix
    return result.rename({'__temp_dim__': dim})


# --------------------------------------------------------------------------- #
#                            pchip interpolation                              #
# --------------------------------------------------------------------------- #

@lazy_guvectorize([
    (int_[:], int_[:], double[:]),
    (double[:], int_[:], double[:]),
    (int_[:], double[:], double[:]),
    (double[:], double[:], double[:]),
], '(n),(n)->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_pchip(x, y, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, num_nan, mask = xynm

    # interpolating function
    ifn = interpolate.PchipInterpolator(x, y, extrapolate=False)
    yf = ifn(x)
    postprocess_nan_func(yf, num_nan, mask, out)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], double[:]),
    (double[:], int_[:], double[:], double[:]),
    (int_[:], double[:], double[:], double[:]),
    (double[:], double[:], double[:], double[:])
], '(n),(n),(m)->(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_pchip_upscale(x, y, ix, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, _, mask = xynm

    # interpolating function
    ifn = interpolate.PchipInterpolator(x, y, extrapolate=False)
    out[:] = ifn(ix)


def _broadcast_pchip(x, y, ix=100, axis=-1):
    if axis != -1:
        y = y.swapaxes(axis, -1)

    if ix is None:
        return _gufunc_pchip(x, y)

    if isinstance(ix, int):
        # automatic upscale
        ix = np.linspace(x.min(), x.max(), ix)

    return _gufunc_pchip_upscale(x, y, ix)


def xr_interp_pchip(obj, dim, ix=100):
    """Interpolate along axis ``dim`` using :func:`scipy.interpolate.pchip`.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        The object to interpolate.
    dim : str
        The axis to interpolate along.
    ix : int or array
        If int, interpolate to this many points spaced evenly along the range
        of the original data. If array, interpolate to those points directly.

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset

    See Also
    --------
    xr_interp
    """

    input_core_dims = [(dim,), (dim,)]
    args = (obj[dim], obj)

    if ix is None:
        kwargs = {'ix': ix, 'axis': -1}
        output_core_dims = [(dim,)]
        return apply_ufunc(_broadcast_pchip, *args, kwargs=kwargs,
                           input_core_dims=input_core_dims,
                           output_core_dims=output_core_dims)

    if isinstance(ix, int):
        ix = np.linspace(float(obj[dim].min()), float(obj[dim].max()), ix)

    kwargs = {'ix': ix, 'axis': -1}
    output_core_dims = [('__temp_dim__',)]

    result = apply_ufunc(_broadcast_pchip, *args, kwargs=kwargs,
                         input_core_dims=input_core_dims,
                         output_core_dims=output_core_dims)
    result['__temp_dim__'] = ix
    return result.rename({'__temp_dim__': dim})


# --------------------------------------------------------------------------- #
#                           scipy signal filtering                            #
# --------------------------------------------------------------------------- #

@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:], double[:]),
    (double[:], int_[:], int_[:], double[:], double[:]),
    (int_[:], double[:], int_[:], double[:], double[:]),
    (double[:], double[:], int_[:], double[:], double[:]),
], '(n),(n),(),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_filter_wiener(x, y, mysize, noise, out=None):  # pragma: no cover
    # Pre-process
    xynm = preprocess_interp1d_nan_func(x, y, out)
    if xynm is None:  # all nan
        return
    x, y, x_even, y_even, num_nan, mask = xynm

    # filter even data
    yf_even = signal.wiener(y_even, mysize=mysize, noise=noise)

    # Post-process
    postprocess_interp1d_nan_func(x, x_even, yf_even, num_nan, mask, out)


def _broadcast_filter_wiener(x, y, mysize=5, noise=1e-2, axis=-1):
    if axis != -1:
        y = y.swapaxes(axis, -1)
    return _gufunc_filter_wiener(x, y, mysize, noise)


def xr_filter_wiener(obj, dim, mysize=5, noise=1e-2):
    kwargs = {'mysize': mysize, 'noise': noise, 'axis': -1}
    input_core_dims = [(dim,), (dim,)]
    output_core_dims = [(dim,)]
    args = (obj[dim], obj)
    return apply_ufunc(_broadcast_filter_wiener, *args,
                       input_core_dims=input_core_dims,
                       output_core_dims=output_core_dims,
                       kwargs=kwargs)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:], double[:]),
    (double[:], int_[:], int_[:], double[:], double[:]),
    (int_[:], double[:], int_[:], double[:], double[:]),
    (double[:], double[:], int_[:], double[:], double[:]),
], '(n),(n),(),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_filtfilt_butter(x, y, N, Wn, out=None):  # pragma: no cover
    # Pre-process
    xynm = preprocess_interp1d_nan_func(x, y, out)
    if xynm is None:  # all nan
        return
    x, y, x_even, y_even, num_nan, mask = xynm

    # filter function
    b, a = signal.butter(N=N[0], Wn=Wn[0])
    # filter even data
    yf_even = signal.filtfilt(b, a, y_even, method='gust')

    # Post-process
    postprocess_interp1d_nan_func(x, x_even, yf_even, num_nan, mask, out)


def _broadcast_filtfilt_butter(x, y, N=2, Wn=0.4, axis=-1):
    if axis != -1:
        y = y.swapaxes(axis, -1)
    return _gufunc_filtfilt_butter(x, y, N, Wn)


def xr_filtfilt_butter(obj, dim, N=2, Wn=0.4):
    """Filter (with forward and backward pass) data along ``dim`` using
    the butterworth design :py:func:`scipy.signal.butter`.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        The object to apply signal filtering to.
    dim : str
        The dimension to filter along.
    N : int, optional
        The order of the filter.
    Wn : scalar, optional
        Critical frequency.
    """
    kwargs = {'N': N, 'Wn': Wn, 'axis': -1}
    input_core_dims = [(dim,), (dim,)]
    output_core_dims = [(dim,)]
    args = (obj[dim], obj)
    return apply_ufunc(_broadcast_filtfilt_butter, *args,
                       input_core_dims=input_core_dims,
                       output_core_dims=output_core_dims,
                       kwargs=kwargs)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:], double[:]),
    (double[:], int_[:], int_[:], double[:], double[:]),
    (int_[:], double[:], int_[:], double[:], double[:]),
    (double[:], double[:], int_[:], double[:], double[:]),
], '(n),(n),(),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_filtfilt_bessel(x, y, N, Wn, out=None):  # pragma: no cover
    # Pre-process
    xynm = preprocess_interp1d_nan_func(x, y, out)
    if xynm is None:  # all nan
        return
    x, y, x_even, y_even, num_nan, mask = xynm

    # filter function
    b, a = signal.bessel(N=N[0], Wn=Wn[0])
    # filter even data
    yf_even = signal.filtfilt(b, a, y_even, method='gust')

    # Post-process
    postprocess_interp1d_nan_func(x, x_even, yf_even, num_nan, mask, out)


def _broadcast_filtfilt_bessel(x, y, N=2, Wn=0.4, axis=-1):
    if axis != -1:
        y = y.swapaxes(axis, -1)
    return _gufunc_filtfilt_bessel(x, y, N, Wn)


def xr_filtfilt_bessel(obj, dim, N=2, Wn=0.4):
    """Filter (with forward and backward pass) data along ``dim`` using
    the bessel design :py:func:`scipy.signal.bessel`.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        The object to apply signal filtering to.
    dim : str
        The dimension to filter along.
    N : int, optional
        The order of the filter.
    Wn : scalar, optional
        Critical frequency.
    """
    kwargs = {'N': N, 'Wn': Wn, 'axis': -1}
    input_core_dims = [(dim,), (dim,)]
    output_core_dims = [(dim,)]
    args = (obj[dim], obj)
    return apply_ufunc(_broadcast_filtfilt_bessel, *args,
                       input_core_dims=input_core_dims,
                       output_core_dims=output_core_dims,
                       kwargs=kwargs)


# --------------------------------------------------------------------------- #
#                               idxmax idxmin                                 #
# --------------------------------------------------------------------------- #

def _index_from_1d_array(array, indices):
    return array.take(indices)


def gufunc_idxmax(x, y, axis=None):
    import dask.array as da
    indx = x.argmax(axis=axis)
    func = functools.partial(_index_from_1d_array, y)

    if isinstance(x, da.Array):
        return da.map_blocks(func, indx, dtype=indx.dtype)
    else:
        return func(indx)


def xr_idxmax(obj, dim):
    """Find the coordinate of the maximum along ``dim``.

    Parameters
    ----------
    obj : xarray.DataArray or xarray.Dataset
        Object to find coordnate maximum in.
    dim : str
        Dimension along which to find maximum

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset
    """
    input_core_dims = [(dim,), (dim,)]
    kwargs = {'axis': -1}
    allna = obj.isnull().all(dim)
    return apply_ufunc(gufunc_idxmax, obj.fillna(-np.inf), obj[dim],
                       input_core_dims=input_core_dims, kwargs=kwargs,
                       dask='allowed').where(~allna)


xr.DataArray.idxmax = xr_idxmax
xr.Dataset.idxmax = xr_idxmax


def gufunc_idxmin(x, y, axis=None):
    import dask.array as da
    indx = x.argmin(axis=axis)
    func = functools.partial(_index_from_1d_array, y)

    if isinstance(x, da.Array):
        return da.map_blocks(func, indx, dtype=indx.dtype)
    else:
        return func(indx)


def xr_idxmin(obj, dim):
    """Find the coordinate of the minimum along ``dim``.

    Parameters
    ----------
    obj : xarray.DataArray or xarray.Dataset
        Object to find coordnate maximum in.
    dim : str
        Dimension along which to find maximum

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset
    """
    input_core_dims = [(dim,), (dim,)]
    kwargs = {'axis': -1}
    allna = obj.isnull().all(dim)
    return apply_ufunc(gufunc_idxmin, obj.fillna(np.inf), obj[dim],
                       input_core_dims=input_core_dims, kwargs=kwargs,
                       dask='allowed').where(~allna)


xr.DataArray.idxmin = xr_idxmin
xr.Dataset.idxmin = xr_idxmin


# --------------------------------------------------------------------------- #
#                      Univariate spline interpolation                        #
# --------------------------------------------------------------------------- #

@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:]),
], '(n),(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_unispline_err(x, y, err, num_knots, out=None):  # pragma: no cover
    xi = x.min()
    xf = x.max()
    t = np.linspace(xi, xf, num_knots)[1:-1]
    fn_interp = LSQUnivariateSpline(x, y, t=t, w=1 / err)
    out[:] = fn_interp(x)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:]),
    (double[:], int_[:], int_[:], double[:]),
    (int_[:], double[:], int_[:], double[:]),
    (double[:], double[:], int_[:], double[:]),
], '(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_unispline_noerr(x, y, num_knots, out=None):  # pragma: no cover
    xi = x.min()
    xf = x.max()
    t = np.linspace(xi, xf, num_knots)[1:-1]
    fn_interp = LSQUnivariateSpline(x, y, t=t)
    out[:] = fn_interp(x)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:], double[:]),
], '(n),(n),(n),(),(m),(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_unispline_err_upscale(x, y, err, num_knots,
                                  ix, out=None):  # pragma: no cover
    xi = x.min()
    xf = x.max()
    t = np.linspace(xi, xf, num_knots)[1:-1]
    fn_interp = LSQUnivariateSpline(x, y, t=t, w=1 / err)
    out[:] = fn_interp(ix)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:], double[:]),
    (double[:], int_[:], int_[:], double[:], double[:]),
    (int_[:], double[:], int_[:], double[:], double[:]),
    (double[:], double[:], int_[:], double[:], double[:]),
], '(n),(n),(),(m),(m)', cache=_NUMBA_CACHE_DEFAULT, forceobj=True)
def _gufunc_unispline_noerr_upscale(x, y, num_knots,
                                    ix, out=None):  # pragma: no cover
    xi = x.min()
    xf = x.max()
    t = np.linspace(xi, xf, num_knots)[1:-1]
    fn_interp = LSQUnivariateSpline(x, y, t=t)
    out[:] = fn_interp(ix)


def _broadcast_unispline(x, y, err=None, num_knots=11, ix=None, axis=-1):
    """Dispatch to the correct numba gufunc.
    """
    if axis != -1:
        y = y.swapaxes(axis, -1)

    if ix is None:
        if err is None:
            return _gufunc_unispline_noerr(x, y, num_knots)
        return _gufunc_unispline_err(x, y, err, num_knots)

    # prepare interpolaing array to pass in
    if isinstance(ix, int):
        ix = np.linspace(x.min(), x.max(), ix)

    # prepare output data to pass in
    if axis < 0:
        axis += y.ndim
    out_shape = y.shape[:-1] + ix.shape
    out = np.empty(out_shape, dtype=x.dtype)

    if err is None:
        _gufunc_unispline_noerr_upscale(x, y, num_knots, ix, out)
    else:
        _gufunc_unispline_err_upscale(x, y, err, num_knots, ix, out)
    return out


def xr_unispline(obj, dim, err=None, num_knots=11, ix=None):
    """Fit a univariate spline along a dimension using linearly spaced knots.

    Parameters
    ----------
    obj : Dataset or DataArray
        Object to fit spline to.
    dim : str
        Dimension to fit spline along.
    err : DataArray or str (optional)
        Error in variables, with with to weight spline fitting. If `err`
        is a string, use the corresponding variable found within `obj`.
    num_knots : int (optional)
        Number of linearly spaced interior knots to form spline with,
        defaults to 11.
    ix : array-like or int (optional)
        Which points to evaluate the newly fitted spline. If int, `ix` many
        points will be chosen linearly spaced across the datas range. If
        None, spline will be evaluated at original coordinates.

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset

    See Also
    --------
    xr_polyfit
    """
    if isinstance(err, str):
        err = obj[err]

    if ix is None:
        kwargs = {'num_knots': num_knots, 'axis': -1}
        if err is None:
            input_core_dims = [(dim,), (dim,)]
            output_core_dims = [(dim,)]
            args = (obj[dim], obj)
        else:
            input_core_dims = [(dim,), (dim,), (dim,)]
            output_core_dims = [(dim,)]
            args = (obj[dim], obj, err)
        return apply_ufunc(_broadcast_unispline, *args,
                           input_core_dims=input_core_dims,
                           output_core_dims=output_core_dims,
                           kwargs=kwargs)
    else:
        if isinstance(ix, int):
            ix = np.linspace(float(obj[dim].min()), float(obj[dim].max()), ix)
        kwargs = {'num_knots': num_knots, 'axis': -1, 'ix': ix}

        if err is None:
            input_core_dims = [(dim,), (dim,)]
            output_core_dims = [('__temp_dim__',)]
            args = (obj[dim], obj)
        else:
            input_core_dims = [(dim,), (dim,), (dim,)]
            output_core_dims = [('__temp_dim__',)]
            args = (obj[dim], obj, err)
        result = apply_ufunc(_broadcast_unispline, *args,
                             input_core_dims=input_core_dims,
                             output_core_dims=output_core_dims,
                             kwargs=kwargs)
        result['__temp_dim__'] = ix
        return result.rename({'__temp_dim__': dim})


# --------------------------------------------------------------------------- #
#                            polynomial fitting                               #
# --------------------------------------------------------------------------- #

@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:]),
    (double[:], int_[:], int_[:], double[:]),
    (int_[:], double[:], int_[:], double[:]),
    (double[:], double[:], int_[:], double[:]),
], '(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT, forceobj=True)
def _gufunc_polyfit(x, y, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, num_nan, mask = xynm

    # interpolating function
    c = np.polynomial.polynomial.polyfit(x, y, deg[0])
    yf = np.polynomial.polynomial.polyval(x, c)
    postprocess_nan_func(yf, num_nan, mask, out)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:])
], '(n),(n),(m),()->(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_polyfit_upscale(x, y, ix, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, _, mask = xynm

    # interpolating function
    c = np.polynomial.polynomial.polyfit(x, y, deg[0])
    out[:] = np.polynomial.polynomial.polyval(ix, c)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:]),
    (double[:], int_[:], int_[:], double[:]),
    (int_[:], double[:], int_[:], double[:]),
    (double[:], double[:], int_[:], double[:]),
], '(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_chebfit(x, y, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, num_nan, mask = xynm

    # interpolating function
    c = np.polynomial.chebyshev.chebfit(x, y, deg[0])
    yf = np.polynomial.chebyshev.chebval(x, c)
    postprocess_nan_func(yf, num_nan, mask, out)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:])
], '(n),(n),(m),()->(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_chebfit_upscale(x, y, ix, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, _, mask = xynm

    # interpolating function
    c = np.polynomial.chebyshev.chebfit(x, y, deg[0])
    out[:] = np.polynomial.chebyshev.chebval(ix, c)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:]),
    (double[:], int_[:], int_[:], double[:]),
    (int_[:], double[:], int_[:], double[:]),
    (double[:], double[:], int_[:], double[:]),
], '(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_legfit(x, y, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, num_nan, mask = xynm

    # interpolating function
    c = np.polynomial.legendre.legfit(x, y, deg[0])
    yf = np.polynomial.legendre.legval(x, c)
    postprocess_nan_func(yf, num_nan, mask, out)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:])
], '(n),(n),(m),()->(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_legfit_upscale(x, y, ix, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, _, mask = xynm

    # interpolating function
    c = np.polynomial.legendre.legfit(x, y, deg[0])
    out[:] = np.polynomial.legendre.legval(ix, c)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:]),
    (double[:], int_[:], int_[:], double[:]),
    (int_[:], double[:], int_[:], double[:]),
    (double[:], double[:], int_[:], double[:]),
], '(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_lagfit(x, y, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, num_nan, mask = xynm

    # interpolating function
    c = np.polynomial.laguerre.lagfit(x, y, deg[0])
    yf = np.polynomial.laguerre.lagval(x, c)
    postprocess_nan_func(yf, num_nan, mask, out)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:])
], '(n),(n),(m),()->(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_lagfit_upscale(x, y, ix, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, _, mask = xynm

    # interpolating function
    c = np.polynomial.laguerre.lagfit(x, y, deg[0])
    out[:] = np.polynomial.laguerre.lagval(ix, c)


@lazy_guvectorize([
    (int_[:], int_[:], int_[:], double[:]),
    (double[:], int_[:], int_[:], double[:]),
    (int_[:], double[:], int_[:], double[:]),
    (double[:], double[:], int_[:], double[:]),
], '(n),(n),()->(n)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_hermfit(x, y, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, num_nan, mask = xynm

    # interpolating function
    c = np.polynomial.hermite.hermfit(x, y, deg[0])
    yf = np.polynomial.hermite.hermval(x, c)
    postprocess_nan_func(yf, num_nan, mask, out)


@lazy_guvectorize([
    (int_[:], int_[:], double[:], int_[:], double[:]),
    (double[:], int_[:], double[:], int_[:], double[:]),
    (int_[:], double[:], double[:], int_[:], double[:]),
    (double[:], double[:], double[:], int_[:], double[:])
], '(n),(n),(m),()->(m)', cache=_NUMBA_CACHE_DEFAULT)
def _gufunc_hermfit_upscale(x, y, ix, deg, out=None):  # pragma: no cover
    xynm = preprocess_nan_func(x, y, out)
    if xynm is None:
        return
    x, y, _, mask = xynm

    # interpolating function
    c = np.polynomial.hermite.hermfit(x, y, deg[0])
    out[:] = np.polynomial.hermite.hermval(ix, c)


POLY_FNS = {
    'polynomial': (_gufunc_polyfit, _gufunc_polyfit_upscale),
    'chebyshev': (_gufunc_chebfit, _gufunc_chebfit_upscale),
    'legendre': (_gufunc_legfit, _gufunc_legfit_upscale),
    'laguerre': (_gufunc_lagfit, _gufunc_lagfit_upscale),
    'hermite': (_gufunc_hermfit, _gufunc_hermfit_upscale),
}


def _broadcast_polyfit(x, y, ix=None, deg=0.5, poly='hermite', axis=-1):
    """Parse arguments and dispatch to the correct function.
    """
    if axis != -1:
        y = y.swapaxes(axis, -1)

    if isinstance(deg, float):
        deg = int(deg * x.size)

    gfn, gfn_upscale = POLY_FNS[poly]

    if ix is None:
        return gfn(x, y, deg)

    if isinstance(ix, int):
        # automatic upscale
        ix = np.linspace(x.min(), x.max(), ix)

    return gfn_upscale(x, y, ix, deg)


def xr_polyfit(obj, dim, ix=None, deg=0.5, poly='hermite'):
    """Fit a polynomial of degree ``deg`` using least-squares along ``dim``.

    Parameters
    ----------
    obj : xarray.Dataset or xarray.DataArray
        The object to fit.
    dim : str, optional
        The dimension to fit along.
    ix : {None, int, array_like}, optional
        If ``None``, interpolate the polynomial at the original x points.
        If ``int``, linearly space this many points along the range of the
        original data and interpolate with these.
        If array-like, interpolate at these given points.
    deg : int or float, optional
        The degree of the polynomial to fit. Used directly if integer. If float
        supplied, with ``0.0 < deg < 1.0``, the proportion of the total
        possible degree to use.
    poly : {'chebyshev', 'polynomial', 'legendre',
            'laguerre', hermite}, optional
        The type of polynomial to fit.

    Returns
    -------
    new_xobj : xarray.DataArray or xarray.Dataset

    See Also
    --------
    xr_unispline
    """
    input_core_dims = [(dim,), (dim,)]
    args = (obj[dim], obj)

    if ix is None:
        kwargs = {'ix': ix, 'axis': -1, 'deg': deg, 'poly': poly}
        output_core_dims = [(dim,)]
        return apply_ufunc(_broadcast_polyfit, *args, kwargs=kwargs,
                           input_core_dims=input_core_dims,
                           output_core_dims=output_core_dims)

    if isinstance(ix, int):
        ix = np.linspace(float(obj[dim].min()), float(obj[dim].max()), ix)

    kwargs = {'ix': ix, 'axis': -1, 'deg': deg, 'poly': poly}
    output_core_dims = [('__temp_dim__',)]

    result = apply_ufunc(_broadcast_polyfit, *args, kwargs=kwargs,
                         input_core_dims=input_core_dims,
                         output_core_dims=output_core_dims)
    result['__temp_dim__'] = ix
    return result.rename({'__temp_dim__': dim})



""" Manage datasets --- loading, saving, merging etc. """

# TODO: add singlet dimensions (for all or given vars) ---------------------- #

import os
from glob import glob
import numpy as np
import xarray as xr


_DEFAULT_FN_CACHE_PATH = '__xyz_cache__'


def cache_to_disk(fn=None, *, cachedir=_DEFAULT_FN_CACHE_PATH, **kwargs):
    """Cache this function to disk, using joblib.
    """
    import joblib
    mem = joblib.Memory(cachedir=cachedir, verbose=0, **kwargs)

    # bare decorator
    if fn:
        return mem.cache(fn)

    # decorator called with kwargs
    else:
        def cache_to_disk_decorator(fn):
            return mem.cache(fn)
        return cache_to_disk_decorator


def _auto_add_extension(file_name, engine):
    """Make sure a file name has an extension that reflects its
    file type.
    """
    if (".h5" not in file_name) and (".nc" not in file_name):
        extension = ".h5" if engine == "h5netcdf" else ".nc"
        file_name += extension
    return file_name


def save_ds(ds, file_name, engine="h5netcdf"):
    """Saves a xarray dataset.

    Parameters
    ----------
        ds: Dataset to save
        file_name: name of file to save to
        engine: engine used to save file

    Returns
    -------
        None
    """
    file_name = _auto_add_extension(file_name, engine)

    # Parse out 'bad' netcdf types (only attributes -> values not so important)
    for attr, val in ds.attrs.items():
        if val is None:
            ds.attrs[attr] = "None"
        if val is True:
            ds.attrs[attr] = "True"
        if val is False:
            ds.attrs[attr] = "False"

    ds.to_netcdf(file_name, engine=engine)


def load_ds(file_name,
            engine="h5netcdf",
            load_to_mem=None,
            create_new=False,
            chunks=None):
    """Loads a xarray dataset. Basically ``xarray.open_dataset`` with some
    different defaults and convenient behaviour.

    Parameters
    ----------
    file_name : str
        Name of file to open.
    engine : str, optional
        Engine used to load file.
    load_to_mem : bool, optional
        Ince opened, load from disk into memory. Defaults to ``True`` if
        ``chunks=None``.
    create_new : bool, optional
        If no file exists make a blank one.
    chunks : int or dict
        Passed to ``xarray.open_dataset`` so that data is stored using
        ``dask.array``.


    Returns
    -------
    ds : xarray.Dataset
        Loaded Dataset.
    """
    file_name = _auto_add_extension(file_name, engine)

    if (load_to_mem is None) and (chunks is None):
        load_to_mem = True
    else:
        if load_to_mem and (chunks is not None):
            raise ValueError("``chunks`` redundant if ``load_to_mem`` given.")
        else:
            load_to_mem = False

    try:
        try:
            ds = xr.open_dataset(file_name,
                                 engine=engine,
                                 chunks=chunks)

        except AttributeError as e1:
            if "object has no attribute" in str(e1) and engine == 'h5netcdf':
                ds = xr.open_dataset(file_name,
                                     engine="netcdf4",
                                     chunks=chunks)
            else:
                raise e1

        if load_to_mem:
            ds.load()
            ds.close()

    except (RuntimeError, OSError) as e2:
        if "o such" in str(e2) and create_new:
            ds = xr.Dataset()
        else:
            raise e2
    return ds


def trimna(obj):
    """Drop values across all dimensions for which all values are NaN.
    """
    trimmed_obj = obj.copy()
    for d in obj.dims:
        trimmed_obj = trimmed_obj.dropna(d, how='all')
    return trimmed_obj


def sort_dims(ds):
    """Make sure all the dimensions on all the variables appear in the
    same order.
    """
    dim_list = tuple(ds.dims)

    for var in ds.data_vars:
        var_dims = (d for d in dim_list if d in ds[var].dims)
        ds[var] = ds[var].transpose(*var_dims)


def post_fix(ds, postfix):
    """Add ``postfix`` to the name of every data variable in ``ds``.
    """
    return ds.rename({old: old + '_' + postfix for old in ds.data_vars})


def check_runs(obj, dim='run', var=None, sel=()):
    """Print out information about the range and any missing values for an
    integer dimension.

    Parameters
    ----------
        obj : xarray object
            Data to check.
        dim : str (optional)
            Dimension to check, defaults to 'run'.
        var : str (optional)
            Subselect this data variable first.
        sel : mapping (optional)
            Subselect these other coordinates first.
    """
    try:
        if sel:
            obj = obj.loc[sel]
        if var:
            obj = obj[var]

        obj = trimna(obj)
        obj = obj.dropna(dim, how='all')
        obj = obj[dim].values
    except KeyError:
        print("* NO DATA *")
        return

    if obj.size == 0:
        print("* NO DATA *")
        return

    if 'int' not in str(obj.dtype):
        raise TypeError("check_runs can only check integer dimesions.")

    xmin, xmax = obj.min(), obj.max() + 1
    xmiss_start = obj[:-1][obj[1:] != obj[:-1] + 1] + 1
    xmiss_end = obj[1:][obj[1:] != obj[:-1] + 1]
    xmissing = list(zip(xmiss_start, xmiss_end))
    msg = "RANGE:  {:>7} -> {:<7} | TOTAL:  {:^7}".format(xmin, xmax, len(obj))
    if xmissing:
        msg += " | MISSING: {}".format(xmissing)
    print(msg)


def auto_xyz_ds(x, y_z=None):
    """Automatically turn an array into a `xarray` dataset. Transpose ``y_z``
    if necessary to automatically match dimension sizes.

    Parameters
    ----------
    x : array_like
        The x-coordinates.
    y_z : array_like, optional
        The y-data, possibly varying with coordinate z.
    """
    # assume non-coordinated - e.g. for auto_histogram
    if y_z is None:
        dims = ['yzwvu'[i] for i in range(x.ndim)]
        return xr.Dataset(data_vars={'x': (dims, x)})

    # Infer dimensions to coords mapping
    y_z = np.array(np.squeeze(y_z), ndmin=2)
    x = np.asarray(x)
    if np.size(x) == y_z.shape[0]:
        y_z = np.transpose(y_z)
    n_y = y_z.shape[0]

    # Turn into dataset
    if x.ndim == 2:
        ds = xr.Dataset(data_vars={'y': (['z', '_x'], y_z),
                                   'x': (['z', '_x'], x)})
    else:
        ds = xr.Dataset(coords={'x': x, 'z': np.arange(n_y)},
                        data_vars={'y': (['z', 'x'], y_z)})
    return ds


def merge_sync_conflict_datasets(base_name, engine='h5netcdf',
                                 combine_first=False):
    """Glob files based on `base_name`, merge them, save this new dataset if
    it contains new info, then clean up the conflicts.

    Parameters
    ----------
        base_name : str
            Base file name to glob on - should include '*'.
        engine : str , optional
            Load and save engine used by xarray.
    """
    fnames = glob(base_name)
    if len(fnames) < 2:
        print('Nothing to do - need multiple files to merge.')
        return

    # make sure first filename is the shortest -> assumed original
    fnames.sort(key=len)

    print("Merging:\n{}\ninto ->\n{}\n".format(fnames, fnames[0]))

    def load_dataset(fname):
        return load_ds(fname, engine=engine)

    datasets = list(map(load_dataset, fnames))

    # combine all the conflicts
    if combine_first:
        full_dataset = datasets[0]
        for ds in datasets[1:]:
            full_dataset = full_dataset.combine_first(ds)
    else:
        full_dataset = xr.merge(datasets)

    # save new dataset?
    if full_dataset.identical(datasets[0]):
        # nothing to do
        pass
    else:
        save_ds(full_dataset, fnames[0], engine=engine)

    # clean up conflicts
    for fname in fnames[1:]:
        os.remove(fname)



# TODO: function_str i.e. (a, b, c) -> (x[t], y[t,w], z)


from cytoolz import isiterable


def _str_2_tuple(x):
    """Ensure `x` is at least a 1-tuple of str.
    """
    return (x,) if isinstance(x, str) else tuple(x)


def dictify(x):
    """Ensure `x` is a dict.
    """
    if isinstance(x, dict):
        return x
    elif x:
        return dict(x)
    else:
        return dict()


def _parse_fn_args(fn_args):
    if fn_args is None:
        return None

    return _str_2_tuple(fn_args)


# combo_runner -------------------------------------------------------------- #

def _parse_combos(combos):
    """Turn dicts and single tuples into proper form for combo runners.
    """
    if isinstance(combos, dict):
        combos = tuple(combos.items())
    elif isinstance(combos[0], str):
        combos = (combos,)
    return tuple((arg, list(vals)) for arg, vals in combos)


def _parse_combo_results(results, var_names):
    """
    """
    if var_names is not None and (isinstance(var_names, str) or
                                  len(var_names) == 1):
        results = (results,)
    return results


# case_runner --------------------------------------------------------------- #

def _parse_cases(cases):
    """
    """

    # cases = {'a': 1, 'b': 2, 'c': 3} --> ({'a': 1, 'b': 2, 'c': 3},)
    if isinstance(cases, dict):
        return (cases,)

    cases = tuple(cases)
    # e.g. if fn_args = ('a',) and cases = (1, 10, 100)
    #     we want cases --> ((1,), (10,), (100,))
    if isinstance(cases[0], str) or not isiterable(cases[0]):
        cases = tuple((c,) for c in cases)

    return cases


def _parse_case_results(results, var_names):
    """
    """
    if isinstance(var_names, str) or len(var_names) == 1:
        results = tuple((r,) for r in results)
    return results


# common variable description ----------------------------------------------- #

def _parse_var_names(var_names):
    """
    """
    return ((None,) if var_names is None else
            (var_names,) if isinstance(var_names, str) else
            tuple(var_names))


def _parse_var_dims(var_dims, var_names):
    """Parse function mapping parameters into standard form.

    Parameters
    ----------
    var_dims : dict, tuple, or str
        * dict
            Mapping of each output to its dimensions, each either str
            or tuple of str. The keys themselves can be a tuple of several
            output names if they all have the same dimensions.
        * tuple
            List of output dimensions directly corresponding to list of
            var_names. Must be same length as `var_names`
        * str
            Only allowed for single output with single dimension.
    var_names : tuple of str, str, or None
        * tuple of str
            List of names of var_names.
        * str
            Single named output.
        * None
            Automatic result output using Dataset/DataArray, in this case
            check that var_dims is None as well.
    """
    if var_names is None:
        if var_dims is not None:
            raise ValueError("Cannot specify variable dimensions if using"
                             "automatic dataset output (var_names=None).")
        return dict()

    new_var_dims = {k: () for k in var_names}  # default to empty tuple

    if not var_dims:
        return new_var_dims
    # check if single output with single dim
    elif isinstance(var_dims, str):
        if len(var_names) != 1:
            raise ValueError("When `var_dims` is specified as a single "
                             "string, there must be a single output, for "
                             "which it is the single dimension.")
        var_dims = {var_names[0]: (var_dims,)}
    # check for direct correspondence to var_names
    elif (isinstance(var_dims, (tuple, list)) and
          any(isinstance(x, str) or x[0] not in var_names for x in var_dims)):
        if len(var_dims) != len(var_names):
            raise ValueError("`var_dims` cannot be interpreted as a "
                             "mapping of `var_names` to their dimensions "
                             "and is the wrong length to be in one to "
                             "one correspondence.")
        var_dims = dict(zip(var_names, var_dims))
    # assume dict-like
    else:
        var_dims = dict(var_dims)

    # update new_var_dims, splitting var_names defined as having the same
    #   dims, and making sure all dims are tuple.
    try:
        for k, v in var_dims.items():
            v = (v,) if isinstance(v, str) else tuple(v)

            if isinstance(k, str):
                if k not in var_names:
                    raise KeyError

                new_var_dims[k] = v

            else:
                for sub_k in k:
                    if sub_k not in var_names:
                        raise KeyError

                    new_var_dims[sub_k] = v

    except KeyError:
        raise ValueError("An unexpected output name was specified in the "
                         "output dimensions mapping.")

    return new_var_dims


_parse_var_coords = dictify
_parse_constants = dictify
_parse_resources = dictify
_parse_attrs = dictify



"""
"""
# TODO: marker alpha

import functools
import itertools
import numpy as np

from ..manage import auto_xyz_ds
from .core import (
    Plotter,
    AbstractLinePlot,
    AbstractScatter,
    AbstractHeatMap,
    calc_row_col_datasets,
    PLOTTER_DEFAULTS,
    intercept_call_arg,
    prettify,
)


@functools.lru_cache(1)
def _init_bokeh_nb():
    """Cache this so it doesn't happen over and over again.
    """
    from bokeh.plotting import output_notebook
    from bokeh.resources import INLINE
    output_notebook(resources=INLINE)


def bshow(figs, nb=True, interactive=False, **kwargs):
    """
    """
    from bokeh.plotting import show
    if nb:
        _init_bokeh_nb()
        show(figs, notebook_handle=interactive)
    else:
        show(figs)


# --------------------------------------------------------------------------- #
#                     Main lineplot interface for bokeh                       #
# --------------------------------------------------------------------------- #

class PlotterBokeh(Plotter):
    def __init__(self, ds, x, y, z=None, **kwargs):
        """
        """
        # bokeh custom options / defaults
        kwargs['return_fig'] = kwargs.pop('return_fig', False)
        self._interactive = kwargs.pop('interactive', False)

        super().__init__(ds, x, y, z, **kwargs, backend='BOKEH')

    def prepare_axes(self):
        """Make the bokeh plot figure and set options.
        """
        from bokeh.plotting import figure

        if self.add_to_axes is not None:
            self._plot = self.add_to_axes

        else:
            # Currently axes scale type must be set at figure creation?
            self._plot = figure(
                # convert figsize to roughly matplotlib dimensions
                width=int(self.figsize[0] * 80 +
                          (100 if self._use_legend else 0) +
                          (20 if self._ytitle else 0) +
                          (20 if not self.yticklabels_hide else 0)),
                height=int(self.figsize[1] * 80 +
                           (20 if self.title else 0) +
                           (20 if self._xtitle else 0) +
                           (20 if not self.xticklabels_hide else 0)),
                x_axis_type=('log' if self.xlog else 'linear'),
                y_axis_type=('log' if self.ylog else 'linear'),
                y_axis_location=('right' if self.ytitle_right else 'left'),
                title=self.title,
                toolbar_location="above",
                toolbar_sticky=False,
                active_scroll="wheel_zoom",
                logo=None,
            )

    def set_axes_labels(self):
        """Set the labels on the axes.
        """
        if self._xtitle:
            self._plot.xaxis.axis_label = self._xtitle
        if self._ytitle:
            self._plot.yaxis.axis_label = self._ytitle

    def set_axes_range(self):
        """Set the plot ranges of the axes, and the panning limits.
        """
        from bokeh.models import DataRange1d

        self.calc_data_range()

        # plt_x_centre = (self._data_xmax + self._data_xmin) / 2
        # plt_x_range = self._data_xmax - self._data_xmin
        # xbounds = (plt_x_centre - plt_x_range, plt_x_centre + plt_x_range)
        xbounds = None
        self._plot.x_range = (DataRange1d(start=self._xlims[0],
                                          end=self._xlims[1],
                                          bounds=xbounds) if self._xlims else
                              DataRange1d(bounds=xbounds))

        # plt_y_centre = (self._data_ymax + self._data_ymin) / 2
        # plt_y_range = abs(self._data_ymax - self._data_ymin)
        # ybounds = (plt_y_centre - plt_y_range, plt_y_centre + plt_y_range)
        ybounds = None
        self._plot.y_range = (DataRange1d(start=self._ylims[0],
                                          end=self._ylims[1],
                                          bounds=ybounds) if self._ylims else
                              DataRange1d(bounds=ybounds))

    def set_spans(self):
        """Set custom horizontal and verical line spans.
        """
        from bokeh.models import Span

        span_opts = {
            'level': 'glyph',
            'line_dash': 'dashed',
            'line_color': (127, 127, 127),
            'line_width': self.span_width,
        }

        if self.hlines:
            for hl in self.hlines:
                self._plot.add_layout(Span(
                    location=hl, dimension='width', **span_opts))
        if self.vlines:
            for vl in self.vlines:
                self._plot.add_layout(Span(
                    location=vl, dimension='height', **span_opts))

    def set_gridlines(self):
        """Set whether to use gridlines or not.
        """
        if not self.gridlines:
            self._plot.xgrid.visible = False
            self._plot.ygrid.visible = False
        else:
            self._plot.xgrid.grid_line_dash = self.gridline_style
            self._plot.ygrid.grid_line_dash = self.gridline_style

    def set_tick_marks(self):
        """Set custom locations for the tick marks.
        """
        from bokeh.models import FixedTicker

        if self.xticks:
            self._plot.xaxis[0].ticker = FixedTicker(ticks=self.xticks)
        if self.yticks:
            self._plot.yaxis[0].ticker = FixedTicker(ticks=self.yticks)

        if self.xticklabels_hide:
            self._plot.xaxis.major_label_text_font_size = '0pt'
        if self.yticklabels_hide:
            self._plot.yaxis.major_label_text_font_size = '0pt'

    def set_sources_heatmap(self):
        from bokeh.plotting import ColumnDataSource

        # initialize empty source
        if not hasattr(self, '_source'):
            self._source = ColumnDataSource(data=dict())

        # remove mask from data -> not necessary soon? / convert to nan?
        var = np.ma.getdata(self._heatmap_var)

        self._source.add([var], 'image')
        self._source.add([self._data_xmin], 'x')
        self._source.add([self._data_ymin], 'y')
        self._source.add([self._data_xmax - self._data_xmin], 'dw')
        self._source.add([self._data_ymax - self._data_ymin], 'dh')

    def set_sources(self):
        """Set the source dictionaries to be used by the plotter functions.
        This is seperate to allow interactive updates of the data only.
        """
        from bokeh.plotting import ColumnDataSource

        # check if heatmap
        if hasattr(self, '_heatmap_var'):
            return self.set_sources_heatmap()

        # 'copy' the zlabels iterator into src_zlbs
        self._zlbls, src_zlbs = itertools.tee(self._zlbls)

        # Initialise with empty dicts
        if not hasattr(self, "_sources"):
            self._sources = [ColumnDataSource(dict())
                             for _ in range(len(self._z_vals))]

        # range through all data and update the sources
        for i, (zlabel, data) in enumerate(zip(src_zlbs, self._gen_xy())):
            self._sources[i].add(data['x'], 'x')
            self._sources[i].add(data['y'], 'y')
            self._sources[i].add([zlabel] * len(data['x']), 'z_coo')

            # check for color for scatter plot
            if 'c' in data:
                self._sources[i].add(data['c'], 'c')

            # check if should set y_err as well
            if 'ye' in data:
                y_err_p = data['y'] + data['ye']
                y_err_m = data['y'] - data['ye']
                self._sources[i].add(
                    list(zip(data['x'], data['x'])), 'y_err_xs')
                self._sources[i].add(list(zip(y_err_p, y_err_m)), 'y_err_ys')

            # check if should set x_err as well
            if 'xe' in data:
                x_err_p = data['x'] + data['xe']
                x_err_m = data['x'] - data['xe']
                self._sources[i].add(
                    list(zip(data['y'], data['y'])), 'x_err_ys')
                self._sources[i].add(list(zip(x_err_p, x_err_m)), 'x_err_xs')

    def plot_legend(self, legend_items=None):
        """Add a legend to the plot.
        """
        if self._use_legend:
            from bokeh.models import Legend

            loc = {'best': 'top_left'}.get(self.legend_loc, self.legend_loc)
            where = {None: 'right'}.get(self.legend_where, self.legend_where)

            # might be manually specified, e.g. from multiplot
            if legend_items is None:
                legend_items = self._lgnd_items

            lg = Legend(items=legend_items)
            lg.location = loc
            lg.click_policy = 'hide'
            self._plot.add_layout(lg, where)

            # Don't repeatedly redraw legend
            self._use_legend = False

    def set_mappable(self):
        from bokeh.models import LogColorMapper, LinearColorMapper
        import matplotlib as plt

        mappr_fn = (LogColorMapper if self.colormap_log else LinearColorMapper)
        bokehpalette = [plt.colors.rgb2hex(m) for m in self.cmap(range(256))]

        self.mappable = mappr_fn(palette=bokehpalette,
                                 low=self._zmin, high=self._zmax)

    def plot_colorbar(self):
        if self._use_colorbar:

            where = {None: 'right'}.get(self.legend_where, self.legend_where)

            from bokeh.models import ColorBar, LogTicker, BasicTicker
            ticker = LogTicker if self.colormap_log else BasicTicker
            color_bar = ColorBar(color_mapper=self.mappable, location=(0, 0),
                                 ticker=ticker(desired_num_ticks=6),
                                 title=self._ctitle)
            self._plot.add_layout(color_bar, where)

    def set_tools(self):
        """Set which tools appear for the plot.
        """
        from bokeh.models import HoverTool

        self._plot.add_tools(HoverTool(tooltips=[
            ("({}, {})".format(self.x_coo, self.y_coo
                               if isinstance(self.y_coo, str) else None),
             "(@x, @y)"), (self.z_coo, "@z_coo")]))

    def update(self):
        from bokeh.io import push_notebook
        self.set_sources()
        push_notebook()

    def show(self, **kwargs):
        """Show the produced figure.
        """
        if self.return_fig:
            return self._plot
        bshow(self._plot, **kwargs)
        return self

    def prepare_plot(self):
        self.prepare_axes()
        self.set_axes_labels()
        self.set_axes_range()
        self.set_spans()
        self.set_gridlines()
        self.set_tick_marks()
        self.set_sources()


def bokeh_multi_plot(fn):
    """Decorate a plotting function to plot a grid of values.
    """

    @functools.wraps(fn)
    def multi_plotter(ds, *args, row=None, col=None, link=False, **kwargs):

        if (row is None) and (col is None):
            return fn(ds, *args, **kwargs)

        # Set some global parameters
        p = fn(ds, *args, **kwargs, call=False)
        p.prepare_data_multi_grid()

        kwargs['xlims'] = kwargs.get('xlims', [p._data_xmin, p._data_xmax])
        kwargs['ylims'] = kwargs.get('ylims', [p._data_ymin, p._data_ymax])

        kwargs['vmin'] = kwargs.pop('vmin', p.vmin)
        kwargs['vmax'] = kwargs.pop('vmax', p.vmax)

        # split the dataset into its respective rows and columns
        ds_r_c, nrows, ncols = calc_row_col_datasets(ds, row=row, col=col)

        # intercept figsize as meaning *total* size for whole grid
        figsize = kwargs.pop('figsize', None)
        if figsize is None:
            av_n = (ncols + nrows) / 2
            figsize = (2 * (4 / av_n)**0.5, 2 * (4 / av_n)**0.5)
        else:
            figsize = (figsize[0] / ncols, figsize[1] / nrows)
        kwargs['figsize'] = figsize

        # intercept return_fig for the full grid and other options
        return_fig = kwargs.pop('return_fig', False)

        subplots = {}

        # range through rows and do subplots
        for i, ds_r in enumerate(ds_r_c):
            # range through columns
            for j, sub_ds in enumerate(ds_r):
                skws = {'legend': False, 'colorbar': False}

                # if not last row
                if i != nrows - 1:
                    skws['xticklabels_hide'] = True
                    skws['xtitle'] = ''

                # if not first column
                if j != 0:
                    skws['yticklabels_hide'] = True
                    skws['ytitle'] = ''

                # label each column
                if (i == 0) and (col is not None):
                    col_val = prettify(ds[col].values[j])
                    skws['title'] = "{} = {}".format(col, col_val)
                    fx = 'fontsize_xtitle'
                    skws['fontsize_title'] = kwargs.get(
                        fx, PLOTTER_DEFAULTS[fx])

                # label each row
                if (j == ncols - 1) and (row is not None):
                    skws['ytitle_right'] = True
                    row_val = prettify(ds[row].values[i])
                    skws['ytitle'] = "{} = {}".format(row, row_val)

                subplots[i, j] = fn(sub_ds, *args, return_fig=True, call=False,
                                    **{**kwargs, **skws})

        from bokeh.layouts import gridplot

        plts = [[subplots[i, j]() for j in range(ncols)] for i in range(nrows)]

        # link zooming and panning between all plots
        if link:
            x_range, y_range = plts[0][0].x_range, plts[0][0].y_range
            for i in range(nrows):
                for j in range(ncols):
                    plts[i][j].x_range = x_range
                    plts[i][j].y_range = y_range

        # the main grid
        p._plot = gridplot(plts)

        if p._use_legend or p._use_colorbar:
            from bokeh.models import Legend, GlyphRenderer, Range1d, ColorBar
            from bokeh.layouts import row

            # plot dummy using last sub_ds
            skws = {'title': "", 'legend_loc': 'center_left',
                    'legend_where': 'left'}
            lgren = fn(sub_ds, *args, return_fig=True, **{**kwargs, **skws})

            # remove all but legend, colorbar and glyph renderers
            lgren.renderers = [
                r for r in lgren.renderers
                if isinstance(r, (Legend, GlyphRenderer, ColorBar))
            ]
            lgren.toolbar_location = None
            lgren.outline_line_color = None

            # size it - this is pretty hacky at the moment
            lgren.width = 120
            lgren.height = int(80 * figsize[1] * nrows + 100)
            lgren.x_range = Range1d(0, 0)
            lgren.y_range = Range1d(0, 0)

            # append to the right of the gridplot
            p._plot = row([p._plot, lgren])

        if return_fig:
            return p._plot
        bshow(p._plot)

    return multi_plotter


class ILinePlot(PlotterBokeh, AbstractLinePlot):

    def __init__(self, ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
        super().__init__(ds, x, y, z=z, y_err=y_err, x_err=x_err, **kwargs)

    def plot_lines(self):
        """Plot the data and a corresponding legend.
        """
        self._lgnd_items = []

        for src in self._sources:
            col = next(self._cols)
            zlabel = next(self._zlbls)
            legend_pics = []

            if self.lines:
                line = self._plot.line(
                    'x', 'y',
                    source=src,
                    color=col,
                    line_dash=next(self._lines),
                    line_width=next(self._lws) * 1.5,
                )
                legend_pics.append(line)

            if self.markers:
                marker = next(self._mrkrs)
                m = getattr(self._plot, marker)(
                    'x', 'y',
                    source=src,
                    name=zlabel,
                    color=col,
                    fill_alpha=0.5,
                    line_width=0.5,
                    size=self._markersize,
                )
                legend_pics.append(m)

            # Check if errors specified as well
            if self.y_err:
                err = self._plot.multi_line(
                    xs='y_err_xs', ys='y_err_ys', source=src, color=col,
                    line_width=self.errorbar_linewidth)
                legend_pics.append(err)
            if self.x_err:
                err = self._plot.multi_line(
                    xs='x_err_xs', ys='x_err_ys', source=src, color=col,
                    line_width=self.errorbar_linewidth)
                legend_pics.append(err)

            # Add the names and styles of drawn lines for the legend
            self._lgnd_items.append((zlabel, legend_pics))

    def __call__(self):
        self.prepare_data_single()
        # Bokeh preparation
        self.prepare_plot()
        self.plot_lines()
        self.plot_legend()
        self.plot_colorbar()
        self.set_tools()
        return self.show(interactive=self._interactive)


@bokeh_multi_plot
@intercept_call_arg
def ilineplot(ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
    """From  ``ds`` plot lines of ``y`` as a function of ``x``, optionally for
    varying ``z``. Interactive,

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str or tuple[str]
        Variable(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return ILinePlot(ds, x, y, z, y_err=y_err, x_err=x_err, **kwargs)


class AutoILinePlot(ILinePlot):
    """Interactive raw data multi-line plot.
    """

    def __init__(self, x, y_z, **lineplot_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, 'x', 'y', z='z', **lineplot_opts)


def auto_ilineplot(x, y_z, **lineplot_opts):
    """Auto version of :func:`~xyzpy.ilineplot` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoILinePlot(x, y_z, **lineplot_opts)()


# --------------------------------------------------------------------------- #

class IScatter(PlotterBokeh, AbstractScatter):

    def __init__(self, ds, x, y, z=None, **kwargs):
        super().__init__(ds, x, y, z, **kwargs, markers=True)

    def plot_scatter(self):
        self._lgnd_items = []

        for src in self._sources:
            if 'c' in src.column_names:
                col = {'field': 'c', 'transform': self.mappable}
            else:
                col = next(self._cols)
            marker = next(self._mrkrs)
            zlabel = next(self._zlbls)
            legend_pics = []

            m = getattr(self._plot, marker)(
                'x', 'y',
                source=src,
                name=zlabel,
                color=col,
                fill_alpha=0.5,
                line_width=0.5,
                size=self._markersize,
            )
            legend_pics.append(m)

            # Add the names and styles of drawn markers for the legend
            self._lgnd_items.append((zlabel, legend_pics))

    def __call__(self):
        self.prepare_data_single()
        # Bokeh preparation
        self.prepare_plot()
        self.plot_scatter()
        self.plot_legend()
        self.plot_colorbar()
        self.set_tools()
        return self.show(interactive=self._interactive)


@bokeh_multi_plot
@intercept_call_arg
def iscatter(ds, x, y, z=None, y_err=None, x_err=None, **kwargs):
    """From  ``ds`` plot a scatter of ``y`` against ``x``, optionally for
    varying ``z``. Interactive.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Quantity to plot along the x-axis.
    y : str or tuple[str]
        Quantity(s) to plot along the y-axis. If tuple, plot each of the
        variables - instead of ``z``.
    z : str, optional
        Dimension to plot into the page.
    y_err : str, optional
        Variable to plot as y-error.
    x_err : str, optional
        Variable to plot as x-error.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return IScatter(ds, x, y, z, y_err=y_err, x_err=x_err, **kwargs)


class AutoIScatter(IScatter):

    def __init__(self, x, y_z, **iscatter_opts):
        ds = auto_xyz_ds(x, y_z)
        super().__init__(ds, 'x', 'y', z='z', **iscatter_opts)


def auto_iscatter(x, y_z, **iscatter_opts):
    """Auto version of :func:`~xyzpy.iscatter` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoIScatter(x, y_z, **iscatter_opts)


# --------------------------------------------------------------------------- #


_HEATMAP_ALT_DEFAULTS = (
    ('legend', False),
    ('colorbar', True),
    ('colormap', 'inferno'),
    ('gridlines', False),
    ('padding', 0),
    ('figsize', (5, 5)),  # try to be square, maybe use aspect_ratio??
)


class IHeatMap(PlotterBokeh, AbstractHeatMap):

    def __init__(self, ds, x, y, z, **kwargs):
        # set some heatmap specific options
        for k, default in _HEATMAP_ALT_DEFAULTS:
            if k not in kwargs:
                kwargs[k] = default
        super().__init__(ds, x, y, z, **kwargs)

    def plot_heatmap(self):
        self.calc_color_norm()
        self._plot.image(image='image', x='x', y='y', dw='dw', dh='dh',
                         source=self._source, color_mapper=self.mappable)

    def __call__(self):
        # Core preparation
        self.prepare_data_single()
        # matplotlib preparation
        self.prepare_plot()
        self.plot_heatmap()
        self.plot_colorbar()
        self.set_tools()
        return self.show(interactive=self._interactive)


@bokeh_multi_plot
@intercept_call_arg
def iheatmap(ds, x, y, z, **kwargs):
    """From  ``ds`` plot variable ``z`` as a function of ``x`` and ``y`` using
    a 2D heatmap. Interactive,

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset to plot from.
    x : str
        Dimension to plot along the x-axis.
    y : str
        Dimension to plot along the y-axis.
    z : str, optional
        Variable to plot as colormap.
    row : str, optional
        Dimension to vary over as a function of rows.
    col : str, optional
        Dimension to vary over as a function of columns.
    plot_opts
        See ``xyzpy.plot.core.PLOTTER_DEFAULTS``.
    """
    return IHeatMap(ds, x, y, z, **kwargs)


class AutoIHeatMap(IHeatMap):

    def __init__(self, x, **iheatmap_opts):
        ds = auto_xyz_ds(x)
        super().__init__(ds, 'x', **iheatmap_opts)


def auto_iheatmap(x, **iheatmap_opts):
    """Auto version of :func:`~xyzpy.iheatmap` that accepts array arguments
    by converting them to a ``Dataset`` first.
    """
    return AutoIHeatMap(x, **iheatmap_opts)()
